#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import torch
from ax.adapter.base import DataLoaderConfig
from ax.adapter.discrete import DiscreteAdapter
from ax.adapter.random import RandomAdapter
from ax.adapter.registry import Generators
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.search_space import SearchSpace
from pyre_extensions import assert_is_instance


DEFAULT_TORCH_DEVICE = torch.device("cpu")
DEFAULT_EHVI_BATCH_LIMIT = 5


"""
Module containing functions that generate standard models, such as Sobol,
GP+EI, etc.

Note: a special case here is a composite generator, which requires an
additional ``GenerationStrategy`` and is able to delegate work to multiple models
(for instance, to a random model to generate the first trial, and to an
optimization model for subsequent trials).

"""


def get_sobol(
    search_space: SearchSpace,
    seed: int | None = None,
    deduplicate: bool = False,
    init_position: int = 0,
    scramble: bool = True,
    fallback_to_sample_polytope: bool = False,
) -> RandomAdapter:
    """Instantiates a Sobol sequence quasi-random generator.

    Args:
        search_space: Sobol generator search space.
        kwargs: Custom args for sobol generator.

    Returns:
        RandomAdapter, with SobolGenerator as model.
    """
    return assert_is_instance(
        Generators.SOBOL(
            experiment=Experiment(search_space=search_space),
            seed=seed,
            deduplicate=deduplicate,
            init_position=init_position,
            scramble=scramble,
            fallback_to_sample_polytope=fallback_to_sample_polytope,
        ),
        RandomAdapter,
    )


def get_factorial(search_space: SearchSpace) -> DiscreteAdapter:
    """Instantiates a factorial generator."""
    return assert_is_instance(
        Generators.FACTORIAL(
            search_space=search_space,
            data_loader_config=DataLoaderConfig(fit_out_of_design=True),
        ),
        DiscreteAdapter,
    )


def get_empirical_bayes_thompson(
    experiment: Experiment,
    data: Data,
    search_space: SearchSpace | None = None,
    num_samples: int = 10000,
    min_weight: float | None = None,
    uniform_weights: bool = False,
) -> DiscreteAdapter:
    """Instantiates an empirical Bayes / Thompson sampling model."""
    if data.df.empty:
        raise ValueError("Empirical Bayes Thompson sampler requires non-empty data.")
    return assert_is_instance(
        Generators.EMPIRICAL_BAYES_THOMPSON(
            experiment=experiment,
            data=data,
            search_space=search_space or experiment.search_space,
            num_samples=num_samples,
            min_weight=min_weight,
            uniform_weights=uniform_weights,
            data_loader_config=DataLoaderConfig(fit_out_of_design=True),
        ),
        DiscreteAdapter,
    )


def get_thompson(
    experiment: Experiment,
    data: Data,
    search_space: SearchSpace | None = None,
    num_samples: int = 10000,
    min_weight: float | None = None,
    uniform_weights: bool = False,
) -> DiscreteAdapter:
    """Instantiates a Thompson sampling model."""
    if data.df.empty:
        raise ValueError("Thompson sampler requires non-empty data.")
    return assert_is_instance(
        Generators.THOMPSON(
            experiment=experiment,
            data=data,
            search_space=search_space or experiment.search_space,
            num_samples=num_samples,
            min_weight=min_weight,
            uniform_weights=uniform_weights,
            data_loader_config=DataLoaderConfig(fit_out_of_design=True),
        ),
        DiscreteAdapter,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.api.client import Client
from ax.api.configs import ChoiceParameterConfig, RangeParameterConfig, StorageConfig
from ax.api.types import TOutcome, TParameterization

__all__ = [
    "Client",
    "ChoiceParameterConfig",
    "RangeParameterConfig",
    "StorageConfig",
    "TOutcome",
    "TParameterization",
]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from math import sqrt

import numpy as np
from ax.adapter.base import DataLoaderConfig
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.transforms.standardize_y import StandardizeY
from ax.core.metric import Metric
from ax.core.objective import Objective
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.outcome_constraint import OutcomeConstraint, ScalarizedOutcomeConstraint
from ax.core.types import ComparisonOp
from ax.exceptions.core import DataRequiredError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_experiment_with_observations,
    get_observations_with_invalid_value,
)
from pandas import DataFrame
from pandas.testing import assert_frame_equal


class StandardizeYTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.obsd1 = ObservationData(
            metric_names=["m1", "m2", "m2"],
            means=np.array([1.0, 2.0, 1.0]),
            covariance=np.array([[1.0, 0.2, 0.4], [0.2, 2.0, 0.8], [0.4, 0.8, 3.0]]),
        )
        self.obsd2 = ObservationData(
            metric_names=["m1", "m1", "m2", "m2"],
            means=np.array([1.0, 1.0, 2.0, 1.0]),
            covariance=np.array(
                [
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.2, 0.4],
                    [0.0, 0.2, 2.0, 0.8],
                    [0.0, 0.4, 0.8, 3.0],
                ]
            ),
        )
        obs1 = Observation(features=ObservationFeatures({}), data=self.obsd1)
        obs2 = Observation(features=ObservationFeatures({}), data=self.obsd2)
        self.t = StandardizeY(
            search_space=None,
            observations=[obs1, obs2],
        )
        self.experiment_data = extract_experiment_data(
            experiment=get_experiment_with_observations(
                observations=[  # Same means as above.
                    [1.0, 2.0],
                    [1.0, 1.0],
                    [1.0, 2.0],
                    [1.0, 1.0],
                ],
                sems=[
                    [1.0, sqrt(2.0)],
                    [1.0, sqrt(3.0)],
                    [1.0, sqrt(2.0)],
                    [1.0, sqrt(3.0)],
                ],
            ),
            data_loader_config=DataLoaderConfig(),
        )

    def test_Init(self) -> None:
        self.assertEqual(self.t.Ymean, {"m1": 1.0, "m2": 1.5})
        self.assertEqual(self.t.Ystd, {"m1": 1.0, "m2": sqrt(1 / 3)})
        with self.assertRaises(DataRequiredError):
            StandardizeY(
                search_space=None,
                observations=[],
            )
        # Initialize with experiment data.
        t = StandardizeY(
            search_space=None,
            experiment_data=self.experiment_data,
        )
        self.assertEqual(t.Ymean, {"m1": 1.0, "m2": 1.5})
        self.assertEqual(t.Ystd, {"m1": 1.0, "m2": sqrt(1 / 3)})

    def test_TransformObservations(self) -> None:
        obsd1_t = ObservationData(
            metric_names=["m1", "m2", "m2"],
            means=np.array([0.0, sqrt(3 / 4), -sqrt(3 / 4)]),
            covariance=np.array(
                [
                    [1.0, 0.2 * sqrt(3), 0.4 * sqrt(3)],
                    [0.2 * sqrt(3), 6.0, 2.4],
                    [0.4 * sqrt(3), 2.4, 9.0],
                ],
            ),
        )
        obsd2 = [deepcopy(self.obsd1)]
        obsd2 = self.t._transform_observation_data(obsd2)
        self.assertTrue(osd_allclose(obsd2[0], obsd1_t))
        obsd2 = self.t._untransform_observation_data(obsd2)
        self.assertTrue(osd_allclose(obsd2[0], self.obsd1))

    def test_TransformOptimizationConfig(self) -> None:
        m1 = Metric(name="m1")
        m2 = Metric(name="m2")
        m3 = Metric(name="m3")
        m4 = Metric(name="m4")
        objective = Objective(metric=m3, minimize=False)
        cons: list[OutcomeConstraint] = [
            OutcomeConstraint(
                metric=m4, op=ComparisonOp.GEQ, bound=2.0, relative=False
            ),
        ]

        oc = OptimizationConfig(objective=objective, outcome_constraints=cons)
        with self.assertRaisesRegex(
            DataRequiredError, "`StandardizeY` transform requires constraint metric"
        ):
            oc = self.t.transform_optimization_config(oc, None, None)
        cons = [
            ScalarizedOutcomeConstraint(
                metrics=[m1, m4],
                weights=[0.5, 0.5],
                op=ComparisonOp.LEQ,
                bound=3.5,
                relative=False,
            ),
        ]
        oc = OptimizationConfig(objective=objective, outcome_constraints=cons)
        with self.assertRaisesRegex(
            DataRequiredError, "`StandardizeY` transform requires constraint metric"
        ):
            oc = self.t.transform_optimization_config(oc, None, None)
        cons = [
            OutcomeConstraint(
                metric=m1, op=ComparisonOp.GEQ, bound=2.0, relative=False
            ),
            OutcomeConstraint(
                metric=m2, op=ComparisonOp.LEQ, bound=3.5, relative=False
            ),
            ScalarizedOutcomeConstraint(
                metrics=[m1, m2],
                weights=[0.5, 0.5],
                op=ComparisonOp.LEQ,
                bound=3.5,
                relative=False,
            ),
        ]
        oc = OptimizationConfig(objective=objective, outcome_constraints=cons)
        oc = self.t.transform_optimization_config(oc, None, None)
        cons_t = [
            OutcomeConstraint(
                metric=m1, op=ComparisonOp.GEQ, bound=1.0, relative=False
            ),
            OutcomeConstraint(
                metric=m2,
                op=ComparisonOp.LEQ,
                bound=2.0 * sqrt(3),  # (3.5 - 1.5) / sqrt(1/3)
                relative=False,
            ),
            ScalarizedOutcomeConstraint(
                metrics=[m1, m2],
                weights=[0.5 * 1.0, 0.5 * sqrt(1 / 3)],
                op=ComparisonOp.LEQ,
                bound=2.25,  # 3.5 - (0.5 * 1.0 + 0.5 * 1.5)
                relative=False,
            ),
        ]
        self.assertTrue(oc.outcome_constraints == cons_t)
        self.assertTrue(oc.objective == objective)

        # Check fail with relative
        con = OutcomeConstraint(
            metric=m1, op=ComparisonOp.GEQ, bound=2.0, relative=True
        )
        oc = OptimizationConfig(objective=objective, outcome_constraints=[con])
        with self.assertRaises(ValueError):
            oc = self.t.transform_optimization_config(oc, None, None)

    def test_non_finite_data_raises(self) -> None:
        for invalid_value in [float("nan"), float("inf")]:
            observations = get_observations_with_invalid_value(
                invalid_value=invalid_value
            )
            with self.assertRaisesRegex(
                ValueError, f"Non-finite data found for metric m1: {invalid_value}"
            ):
                StandardizeY(observations=observations, config={"metrics": ["m1"]})

    def test_transform_experiment_data(self) -> None:
        experiment_data = deepcopy(self.experiment_data)
        transformed_data = self.t.transform_experiment_data(
            experiment_data=experiment_data
        )
        # Check that arm data is identical.
        self.assertIs(experiment_data.arm_data, transformed_data.arm_data)
        assert_frame_equal(
            self.experiment_data.arm_data,
            transformed_data.arm_data,
        )
        # Check that observation data is transformed correctly.
        observation_data = transformed_data.observation_data
        sq3 = sqrt(3.0)
        expected_means = DataFrame(
            index=observation_data.index,
            columns=observation_data["mean"].columns,
            data=[
                [0.0, 0.5 * sq3],
                [0.0, -0.5 * sq3],
                [0.0, 0.5 * sq3],
                [0.0, -0.5 * sq3],
            ],
        )
        assert_frame_equal(observation_data["mean"], expected_means)
        expected_sems = DataFrame(
            index=observation_data.index,
            columns=observation_data["sem"].columns,
            data=[
                [1.0, sqrt(2.0) * sq3],
                [1.0, sqrt(3.0) * sq3],
                [1.0, sqrt(2.0) * sq3],
                [1.0, sqrt(3.0) * sq3],
            ],
        )
        assert_frame_equal(observation_data["sem"], expected_sems)


def osd_allclose(osd1: ObservationData, osd2: ObservationData) -> bool:
    if osd1.metric_names != osd2.metric_names:
        return False
    if not np.allclose(osd1.means, osd2.means):
        return False
    if not np.allclose(osd1.covariance, osd2.covariance):
        return False
    return True



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import warnings
from collections import OrderedDict
from unittest.mock import Mock

import numpy as np
import torch
from ax.core.search_space import SearchSpaceDigest
from ax.exceptions.core import AxWarning, UnsupportedError, UserInputError
from ax.generators.torch.botorch_modular.kernels import ScaleMaternKernel
from ax.generators.torch.botorch_modular.utils import (
    _get_shared_rows,
    choose_botorch_acqf_class,
    choose_model_class,
    construct_acquisition_and_optimizer_options,
    convert_to_block_design,
    get_cv_fold,
    logger,
    ModelConfig,
    subset_state_dict,
    use_model_list,
)
from ax.generators.torch.utils import _to_inequality_constraints, predict_from_model
from ax.generators.torch_base import TorchOptConfig
from ax.generators.types import TConfig
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.torch_stubs import get_torch_test_data
from botorch.acquisition import qLogNoisyExpectedImprovement
from botorch.acquisition.analytic import PosteriorMean
from botorch.acquisition.multi_objective.logei import (
    qLogNoisyExpectedHypervolumeImprovement,
)
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.gp_regression import SingleTaskGP
from botorch.models.gp_regression_fidelity import SingleTaskMultiFidelityGP
from botorch.models.gp_regression_mixed import MixedSingleTaskGP
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.models.multitask import MultiTaskGP
from botorch.posteriors.ensemble import EnsemblePosterior
from botorch.utils.datasets import SupervisedDataset
from pyre_extensions import assert_is_instance, none_throws


class BoTorchGeneratorUtilsTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.dtype = torch.float
        (
            self.Xs,
            self.Ys,
            self.Yvars,
            _,
            _,
            self.feature_names,
            self.metric_names,
        ) = get_torch_test_data(dtype=self.dtype)
        self.Xs2, self.Ys2, self.Yvars2, _, _, _, _ = get_torch_test_data(
            dtype=self.dtype,
            offset=1.0,  # Making this data different.
        )
        self.fixed_noise_dataset = SupervisedDataset(
            X=self.Xs,
            Y=self.Ys,
            Yvar=self.Yvars,
            feature_names=self.feature_names,
            outcome_names=self.metric_names,
        )
        self.supervised_dataset = SupervisedDataset(
            X=self.Xs,
            Y=self.Ys,
            feature_names=self.feature_names,
            outcome_names=self.metric_names,
        )
        self.none_Yvars = [torch.tensor([[np.nan], [np.nan]])]
        self.task_features = []
        self.objective_thresholds = torch.tensor([0.5, 1.5])

    def test_choose_model_class_fidelity_features(self) -> None:
        # Only a single fidelity feature can be used.
        with self.assertRaisesRegex(
            NotImplementedError, "Only a single fidelity feature"
        ):
            choose_model_class(
                dataset=self.fixed_noise_dataset,
                search_space_digest=SearchSpaceDigest(
                    feature_names=[], bounds=[], fidelity_features=[1, 2]
                ),
            )
        # No support for non-empty task & fidelity features yet.
        with self.assertRaisesRegex(NotImplementedError, "Multi-task multi-fidelity"):
            choose_model_class(
                dataset=self.fixed_noise_dataset,
                search_space_digest=SearchSpaceDigest(
                    feature_names=[],
                    bounds=[],
                    task_features=[1],
                    fidelity_features=[1],
                ),
            )
        # With fidelity features, use SingleTaskMultiFidelityGP.
        for ds in [self.supervised_dataset, self.fixed_noise_dataset]:
            self.assertEqual(
                SingleTaskMultiFidelityGP,
                choose_model_class(
                    dataset=ds,
                    search_space_digest=SearchSpaceDigest(
                        feature_names=[],
                        bounds=[],
                        fidelity_features=[2],
                    ),
                ),
            )

    def test_choose_model_class_task_features(self) -> None:
        # Only a single task feature can be used.
        with self.assertRaisesRegex(NotImplementedError, "Only a single task feature"):
            choose_model_class(
                dataset=self.fixed_noise_dataset,
                search_space_digest=SearchSpaceDigest(
                    feature_names=[], bounds=[], task_features=[1, 2]
                ),
            )
        # With task features use MultiTaskGP.
        for datasets in (self.supervised_dataset, self.fixed_noise_dataset):
            self.assertEqual(
                MultiTaskGP,
                choose_model_class(
                    dataset=datasets,
                    search_space_digest=SearchSpaceDigest(
                        feature_names=[], bounds=[], task_features=[1]
                    ),
                ),
            )

    def test_choose_model_class_discrete_features(self) -> None:
        # With discrete features, use MixedSingleTaskyGP.
        self.assertEqual(
            MixedSingleTaskGP,
            choose_model_class(
                dataset=self.supervised_dataset,
                search_space_digest=SearchSpaceDigest(
                    feature_names=[],
                    bounds=[],
                    task_features=[],
                    categorical_features=[1],
                ),
            ),
        )

    def test_choose_model_class(self) -> None:
        # Without fidelity/task features, use SingleTaskGP.
        for ds in [self.fixed_noise_dataset, self.supervised_dataset]:
            self.assertEqual(
                SingleTaskGP,
                choose_model_class(
                    dataset=ds,
                    search_space_digest=SearchSpaceDigest(
                        feature_names=[],
                        bounds=[],
                    ),
                ),
            )

    def test_choose_botorch_acqf_class(self) -> None:
        self.assertEqual(
            qLogNoisyExpectedImprovement,
            choose_botorch_acqf_class(
                torch_opt_config=TorchOptConfig(
                    objective_weights=torch.tensor([1.0, 0.0]),
                    is_moo=False,
                ),
                datasets=[self.supervised_dataset],
            ),
        )
        self.assertEqual(
            qLogNoisyExpectedHypervolumeImprovement,
            choose_botorch_acqf_class(
                torch_opt_config=TorchOptConfig(
                    objective_weights=torch.tensor([1.0, -1.0]),
                    is_moo=True,
                ),
                datasets=[self.supervised_dataset],
            ),
        )
        # Check that it doesn't error out when given a mix of known and unknown noise.
        ds1 = self.fixed_noise_dataset
        ds1.outcome_names = ["y2"]
        datasets = [self.supervised_dataset, ds1]
        with self.assertLogs(logger=logger, level="DEBUG") as logs:
            choose_botorch_acqf_class(
                torch_opt_config=TorchOptConfig(
                    objective_weights=torch.tensor([1.0, -1.0]),
                    is_moo=True,
                    outcome_constraints=(
                        torch.tensor([[1.0, 0.0]]),
                        torch.tensor([[4.0]]),
                    ),
                ),
                datasets=datasets,
            )
        self.assertTrue(
            any(
                "Only a subset of datasets have noise observations." in str(log)
                for log in logs
            )
        )

    def test_construct_acquisition_and_optimizer_options(self) -> None:
        # Two dicts for `Acquisition` should be concatenated
        acqf_options = {}
        botorch_acqf_options: TConfig = {Keys.NUM_FANTASIES: 64}

        acquisition_function_kwargs = {Keys.CURRENT_VALUE: torch.tensor([1.0])}
        ax_acquisition_kwargs = {Keys.SUBSET_MODEL: False}
        optimizer_kwargs = {Keys.NUM_RESTARTS: 40, Keys.RAW_SAMPLES: 1024}
        model_gen_options = {
            Keys.AX_ACQUISITION_KWARGS: ax_acquisition_kwargs,
            Keys.ACQF_KWARGS: acquisition_function_kwargs,
            Keys.OPTIMIZER_KWARGS: optimizer_kwargs,
        }

        (
            final_acq_options,
            final_botorch_acqf_options,
            final_opt_options,
            final_botorch_acqf_classes_with_options,
        ) = construct_acquisition_and_optimizer_options(
            acqf_options=acqf_options,
            botorch_acqf_options=botorch_acqf_options,
            # pyre-fixme[6]: Incompatible parameter type [6]:
            # In call `construct_acquisition_and_optimizer_options`, for
            # argument `model_gen_options`, expected `Optional[Dict[str,
            # Union[None, Dict[int, typing.Any], Dict[str, typing.Any],
            # List[int], List[str], OptimizationConfig, WinsorizationConfig,
            # AcquisitionFunction, float, int, str]]]` but got `Dict[Keys,
            # Union[Dict[Keys, bool], Dict[Keys, int], Dict[Keys, Tensor]]]`.
            model_gen_options=model_gen_options,
        )
        self.assertEqual(final_acq_options, {Keys.SUBSET_MODEL: False})
        self.assertEqual(
            final_botorch_acqf_options,
            {Keys.NUM_FANTASIES: 64, Keys.CURRENT_VALUE: torch.tensor([1.0])},
        )
        self.assertEqual(final_opt_options, optimizer_kwargs)
        self.assertIsNone(final_botorch_acqf_classes_with_options)

        with self.assertRaisesRegex(
            ValueError, "Found forbidden keys in `model_gen_options`"
        ):
            construct_acquisition_and_optimizer_options(
                acqf_options=acqf_options,
                botorch_acqf_options=botorch_acqf_options,
                model_gen_options={**model_gen_options, "extra": "key"},
            )

        # test with botorch_acqf_classes_with_options
        botorch_acqf_classes_with_options = [
            (PosteriorMean, {}),
            (qLogNoisyExpectedImprovement, {}),
        ]
        with warnings.catch_warnings(record=True) as ws:
            (
                final_acq_options,
                final_botorch_acqf_options,
                final_opt_options,
                final_botorch_acqf_classes_with_options,
            ) = construct_acquisition_and_optimizer_options(
                acqf_options=acqf_options,
                botorch_acqf_options=botorch_acqf_options,
                # pyre-fixme[6]: Incompatible parameter type [6]:
                # In call `construct_acquisition_and_optimizer_options`, for
                # argument `model_gen_options`, expected `Optional[Dict[str,
                # Union[None, Dict[int, typing.Any], Dict[str, typing.Any],
                # List[int], List[str], OptimizationConfig, WinsorizationConfig,
                # AcquisitionFunction, float, int, str]]]` but got `Dict[Keys,
                # Union[Dict[Keys, bool], Dict[Keys, int], Dict[Keys, Tensor]]]`.
                model_gen_options=model_gen_options,
                # pyre-fixme[6]: Incompatible parameter type [6]: In call
                # `construct_acquisition_and_optimizer_options`, for argument
                # `botorch_acqf_classes_with_options`, expected `Optional[
                # List[Tuple[Type[AcquisitionFunction], Dict[str, Union[None,
                # Dict[int, typing.Any], Dict[str, typing.Any], List[int], List[str],
                # OptimizationConfig, WinsorizationConfig, AcquisitionFunction,
                # float, int, str]]]]]` but got `List[Tuple[Type[
                # qLogNoisyExpectedImprovement], Dict[typing.Any, typing.Any]]]`.
                botorch_acqf_classes_with_options=botorch_acqf_classes_with_options,
            )
            self.assertEqual(
                botorch_acqf_classes_with_options,
                final_botorch_acqf_classes_with_options,
            )
            self.assertEqual(final_botorch_acqf_options, botorch_acqf_options)
            self.assertEqual(final_acq_options, {Keys.SUBSET_MODEL: False})
            self.assertEqual(final_opt_options, optimizer_kwargs)
            self.assertEqual(len(ws), 1)
            warning = ws[0]
            self.assertEqual(warning.category, AxWarning)
            self.assertEqual(
                str(warning.message),
                "botorch_acqf_options are being ignored, due to using "
                "MultiAcquisition. Specify options for each acquistion function"
                "via botorch_acqf_classes_with_options.",
            )

        # test that botorch_acqf_classes_with_options is updated
        botorch_acqf_classes_with_options = [
            (qLogNoisyExpectedImprovement, {Keys.NUM_FANTASIES: 64}),
        ]
        (
            final_acq_options,
            final_botorch_acqf_options,
            final_opt_options,
            final_botorch_acqf_classes_with_options,
        ) = construct_acquisition_and_optimizer_options(
            acqf_options=acqf_options,
            botorch_acqf_options=botorch_acqf_options,
            # pyre-fixme[6]: Incompatible parameter type [6]:
            # In call `construct_acquisition_and_optimizer_options`, for
            # argument `model_gen_options`, expected `Optional[Dict[str,
            # Union[None, Dict[int, typing.Any], Dict[str, typing.Any],
            # List[int], List[str], OptimizationConfig, WinsorizationConfig,
            # AcquisitionFunction, float, int, str]]]` but got `Dict[Keys,
            # Union[Dict[Keys, bool], Dict[Keys, int], Dict[Keys, Tensor]]]`.
            model_gen_options=model_gen_options,
            # pyre-fixme[6]: Incompatible parameter type [6]: In call
            # `construct_acquisition_and_optimizer_options`, for argument
            # `botorch_acqf_classes_with_options`, expected `Optional[
            # List[Tuple[Type[AcquisitionFunction], Dict[str, Union[None,
            # Dict[int, typing.Any], Dict[str, typing.Any], List[int], List[str],
            # OptimizationConfig, WinsorizationConfig, AcquisitionFunction,
            # float, int, str]]]]]` but got `List[Tuple[Type[
            # qLogNoisyExpectedImprovement], Dict[typing.Any, typing.Any]]]`.
            botorch_acqf_classes_with_options=botorch_acqf_classes_with_options,
        )
        self.assertEqual(
            [
                (
                    qLogNoisyExpectedImprovement,
                    {Keys.NUM_FANTASIES: 64, Keys.CURRENT_VALUE: torch.tensor([1.0])},
                )
            ],
            final_botorch_acqf_classes_with_options,
        )
        self.assertEqual(final_botorch_acqf_options, botorch_acqf_options)
        self.assertEqual(final_acq_options, {Keys.SUBSET_MODEL: False})
        self.assertEqual(final_opt_options, optimizer_kwargs)

    def test_use_model_list(self) -> None:
        self.assertFalse(
            use_model_list(
                datasets=[self.supervised_dataset],
                model_configs=[ModelConfig(botorch_model_class=SingleTaskGP)],
                search_space_digest=SearchSpaceDigest(feature_names=[], bounds=[]),
            )
        )
        self.assertFalse(  # Batched multi-output case.
            use_model_list(
                datasets=[
                    SupervisedDataset(
                        X=self.Xs,
                        Y=self.Ys,
                        feature_names=self.feature_names,
                        outcome_names=["y"],
                    )
                ],
                model_configs=[ModelConfig(botorch_model_class=SingleTaskGP)],
                search_space_digest=SearchSpaceDigest(feature_names=[], bounds=[]),
            )
        )
        # Multi-output with allow_batched_models
        self.assertFalse(
            use_model_list(
                datasets=2
                * [
                    SupervisedDataset(
                        X=self.Xs,
                        Y=self.Ys,
                        feature_names=self.feature_names,
                        outcome_names=["y"],
                    )
                ],
                model_configs=[ModelConfig(botorch_model_class=SingleTaskGP)],
                search_space_digest=SearchSpaceDigest(feature_names=[], bounds=[]),
                allow_batched_models=True,
            )
        )
        self.assertTrue(
            use_model_list(
                datasets=2
                * [
                    SupervisedDataset(
                        X=self.Xs,
                        Y=self.Ys,
                        feature_names=self.feature_names,
                        outcome_names=["y"],
                    )
                ],
                model_configs=[ModelConfig(botorch_model_class=SingleTaskGP)],
                search_space_digest=SearchSpaceDigest(feature_names=[], bounds=[]),
                allow_batched_models=False,
            )
        )
        self.assertTrue(
            use_model_list(
                datasets=[
                    SupervisedDataset(
                        X=self.Xs,
                        Y=self.Ys,
                        feature_names=self.feature_names,
                        outcome_names=["y"],
                    ),
                    SupervisedDataset(
                        X=self.Xs2,
                        Y=self.Ys2,
                        feature_names=self.feature_names,
                        outcome_names=["y"],
                    ),
                ],
                search_space_digest=SearchSpaceDigest(feature_names=[], bounds=[]),
                model_configs=[ModelConfig(botorch_model_class=SingleTaskGP)],
            )
        )
        self.assertFalse(
            use_model_list(
                datasets=[self.supervised_dataset],
                model_configs=[ModelConfig(botorch_model_class=MultiTaskGP)],
                search_space_digest=SearchSpaceDigest(
                    feature_names=[], bounds=[], task_features=[1, 2]
                ),
            )
        )
        # Not using model list with single outcome.
        self.assertFalse(
            use_model_list(
                datasets=[self.supervised_dataset],
                model_configs=[
                    ModelConfig(botorch_model_class=SaasFullyBayesianSingleTaskGP)
                ],
                search_space_digest=SearchSpaceDigest(feature_names=[], bounds=[]),
            )
        )
        # Using it with multiple outcomes.
        self.assertTrue(
            use_model_list(
                datasets=[
                    SupervisedDataset(
                        X=self.Xs,
                        Y=self.Ys,
                        feature_names=self.feature_names,
                        outcome_names=["y"],
                    ),
                    SupervisedDataset(
                        X=self.Xs2,
                        Y=self.Ys2,
                        feature_names=self.feature_names,
                        outcome_names=["y"],
                    ),
                ],
                model_configs=[
                    ModelConfig(botorch_model_class=SaasFullyBayesianSingleTaskGP)
                ],
                search_space_digest=SearchSpaceDigest(feature_names=[], bounds=[]),
            )
        )
        self.assertTrue(
            use_model_list(
                datasets=[
                    SupervisedDataset(
                        X=self.Xs,
                        Y=self.Ys.repeat(1, 2),
                        feature_names=self.feature_names,
                        outcome_names=["y1", "y2"],
                    ),
                ],
                model_configs=[
                    ModelConfig(botorch_model_class=SaasFullyBayesianSingleTaskGP)
                ],
                search_space_digest=SearchSpaceDigest(feature_names=[], bounds=[]),
            )
        )

    def test_get_shared_rows(self) -> None:
        # test bad input
        with self.assertRaisesRegex(
            UserInputError, "All inputs must be two-dimensional."
        ):
            _get_shared_rows(Xs=[torch.rand(3, 4, 2), torch.rand(2, 4, 2)])

        X1 = torch.rand(4, 2)

        # X1 is subset of X2
        X2 = torch.cat((X1[:2], torch.rand(1, 2), X1[2:]))
        X_shared, shared_idcs = _get_shared_rows([X1, X2])
        # unique() here just makes sure the order is consistent
        self.assertTrue(torch.equal(X1.unique(dim=0), X_shared))
        self.assertTrue(torch.equal(shared_idcs[0], torch.arange(4)))
        self.assertTrue(torch.equal(shared_idcs[1], torch.tensor([0, 1, 3, 4])))
        # X2 is subset of X1
        X2 = X1[:3]
        X_shared, shared_idcs = _get_shared_rows([X1, X2])
        self.assertTrue(torch.equal(X2.unique(dim=0), X_shared))
        self.assertTrue(torch.equal(shared_idcs[0], torch.arange(3)))
        self.assertTrue(torch.equal(shared_idcs[1], torch.arange(3)))

        # no overlap
        X2 = torch.rand(2, 2)
        X_shared, shared_idcs = _get_shared_rows([X1, X2])
        self.assertEqual(X_shared.numel(), 0)
        self.assertEqual(shared_idcs[0].numel(), 0)
        self.assertEqual(shared_idcs[1].numel(), 0)

        # three tensors
        X2 = torch.cat((X1[:2], torch.rand(1, 2), X1[2:]))
        X3 = torch.cat((torch.rand(1, 2), X1[:2], torch.rand(1, 2), X1[3:4]))
        X_shared, shared_idcs = _get_shared_rows([X1, X2, X3])
        self.assertTrue(torch.equal(shared_idcs[0], torch.tensor([0, 1, 3])))
        self.assertTrue(torch.equal(shared_idcs[1], torch.tensor([0, 1, 4])))
        self.assertTrue(
            torch.equal(X_shared, X1[torch.tensor([0, 1, 3])].unique(dim=0))
        )

        # test tensors with duplicate rows
        X4 = torch.cat((X1[:2], X1[:2]))
        X_shared, shared_idcs = _get_shared_rows(Xs=[X1, X4])
        self.assertTrue(torch.equal(X_shared, X1[:2].unique(dim=0)))
        self.assertTrue(torch.equal(shared_idcs[0], torch.tensor([0, 1])))
        self.assertTrue(torch.equal(shared_idcs[1], torch.tensor([2, 3])))

    def test_convert_to_block_design(self) -> None:
        # simple case: block design, supervised
        X = torch.rand(4, 2)
        Ys = [torch.rand(4, 1), torch.rand(4, 1)]
        metric_names = ["y1", "y2"]
        datasets_unknown_noise = [
            SupervisedDataset(
                X=X,
                Y=Ys[i],
                feature_names=["x1", "x2"],
                outcome_names=[metric_names[i]],
            )
            for i in range(2)
        ]
        new_datasets = convert_to_block_design(datasets=datasets_unknown_noise)
        self.assertEqual(len(new_datasets), 1)
        self.assertIsInstance(new_datasets[0], SupervisedDataset)
        self.assertTrue(torch.equal(new_datasets[0].X, X))
        self.assertTrue(torch.equal(new_datasets[0].Y, torch.cat(Ys, dim=-1)))
        self.assertEqual(new_datasets[0].outcome_names, metric_names)

        # simple case: block design, fixed
        Yvars = [torch.rand(4, 1), torch.rand(4, 1)]
        datasets_noisy = [
            SupervisedDataset(
                X=X,
                Y=Ys[i],
                Yvar=Yvars[i],
                feature_names=["x1", "x2"],
                outcome_names=[metric_names[i]],
            )
            for i in range(2)
        ]
        new_datasets = convert_to_block_design(datasets=datasets_noisy)
        self.assertEqual(len(new_datasets), 1)
        self.assertIsNotNone(new_datasets[0].Yvar)
        self.assertTrue(torch.equal(new_datasets[0].X, X))
        self.assertTrue(torch.equal(new_datasets[0].Y, torch.cat(Ys, dim=-1)))
        self.assertTrue(
            torch.equal(none_throws(new_datasets[0].Yvar), torch.cat(Yvars, dim=-1))
        )
        self.assertEqual(new_datasets[0].outcome_names, metric_names)

        # test error is raised if not block design and force=False
        X2 = torch.cat((X[:3], torch.rand(1, 2)))
        datasets = [
            SupervisedDataset(
                X=X, Y=Y, feature_names=["x1", "x2"], outcome_names=[name]
            )
            for X, Y, name in zip((X, X2), Ys, metric_names)
        ]
        with self.assertRaisesRegex(
            UnsupportedError, "Cannot convert data to non-block design data."
        ):
            convert_to_block_design(datasets=datasets)

        # test a log is produced if not block design and force=True (supervised)
        with self.assertLogs(logger=logger, level="DEBUG") as logs:
            new_datasets = convert_to_block_design(datasets=datasets, force=True)
        self.assertTrue(
            any(
                "Forcing conversion of data not complying to a block design" in str(log)
                for log in logs
            )
        )
        self.assertEqual(len(new_datasets), 1)
        self.assertIsNone(new_datasets[0].Yvar)
        # unique() here makes sure the order is consistent
        self.assertTrue(torch.equal(new_datasets[0].X, X[:3].unique(dim=0)))
        self.assertTrue(
            torch.equal(new_datasets[0].Y, torch.cat([Y[:3] for Y in Ys], dim=-1))
        )
        self.assertEqual(new_datasets[0].outcome_names, metric_names)

        # test a log is produced if not block design and force=True (fixed)
        datasets = [
            SupervisedDataset(
                X=X, Y=Y, Yvar=Yvar, feature_names=["x1", "x2"], outcome_names=[name]
            )
            for X, Y, Yvar, name in zip((X, X2), Ys, Yvars, metric_names)
        ]
        with self.assertLogs(logger=logger, level="DEBUG") as logs:
            new_datasets = convert_to_block_design(datasets=datasets, force=True)
        self.assertTrue(
            any(
                "Forcing conversion of data not complying to a block design" in str(log)
                for log in logs
            )
        )
        self.assertEqual(len(new_datasets), 1)
        self.assertIsNotNone(new_datasets[0].Yvar)
        self.assertTrue(torch.equal(new_datasets[0].X, X[:3].unique(dim=0)))
        self.assertTrue(
            torch.equal(new_datasets[0].Y, torch.cat([Y[:3] for Y in Ys], dim=-1))
        )
        self.assertTrue(
            torch.equal(
                none_throws(new_datasets[0].Yvar),
                torch.cat([Yvar[:3] for Yvar in Yvars], dim=-1),
            )
        )
        self.assertEqual(new_datasets[0].outcome_names, metric_names)

        # Test that known and unknown noise can be merged if force=True.
        datasets = [datasets_unknown_noise[0], datasets_noisy[1]]
        with self.assertLogs(logger=logger, level="DEBUG") as logs:
            new_datasets = convert_to_block_design(datasets=datasets, force=True)
        self.assertTrue(
            any(
                "Only a subset of datasets have noise observations." in str(log)
                for log in logs
            )
        )
        self.assertEqual(len(new_datasets), 1)
        self.assertIsNone(new_datasets[0].Yvar)
        # Errors out if force=False.
        with self.assertRaisesRegex(
            UnsupportedError, "Cannot convert mixed data with and without variance"
        ):
            convert_to_block_design(datasets=datasets, force=False)

    def test_to_inequality_constraints(self) -> None:
        A = torch.tensor([[0, 1, -2, 3], [0, 1, 0, 0]])
        b = torch.tensor([[1], [2]])
        ineq_constraints = none_throws(
            _to_inequality_constraints(linear_constraints=(A, b))
        )
        self.assertEqual(len(ineq_constraints), 2)
        self.assertAllClose(ineq_constraints[0][0], torch.tensor([1, 2, 3]))
        self.assertAllClose(ineq_constraints[0][1], torch.tensor([-1, 2, -3]))
        self.assertEqual(ineq_constraints[0][2], -1.0)
        self.assertAllClose(ineq_constraints[1][0], torch.tensor([1]))
        self.assertAllClose(ineq_constraints[1][1], torch.tensor([-1]))
        self.assertEqual(ineq_constraints[1][2], -2.0)

    def test_subset_state_dict(self) -> None:
        m0 = SingleTaskGP(train_X=torch.rand(5, 2), train_Y=torch.rand(5, 1))
        m1 = SingleTaskGP(train_X=torch.rand(5, 2), train_Y=torch.rand(5, 1))
        model_list = ModelListGP(m0, m1)
        model_list_state_dict = assert_is_instance(model_list.state_dict(), OrderedDict)
        # Subset the model dict from model list and check that it is correct.
        m0_state_dict = model_list.models[0].state_dict()
        subsetted_m0_state_dict = subset_state_dict(
            state_dict=model_list_state_dict, submodel_index=0
        )
        self.assertEqual(m0_state_dict.keys(), subsetted_m0_state_dict.keys())
        for k in m0_state_dict:
            self.assertTrue(torch.equal(m0_state_dict[k], subsetted_m0_state_dict[k]))
        # Check that it can be loaded on the model.
        m0.load_state_dict(subsetted_m0_state_dict)

    def test_get_folds(self) -> None:
        X = torch.rand(10, 2)
        Y = torch.rand(10, 1)
        Yvar = torch.rand(10, 1)
        dataset = SupervisedDataset(
            X=X, Y=Y, Yvar=Yvar, feature_names=["x1", "x2"], outcome_names=["y"]
        )
        # CV
        cv_fold = get_cv_fold(
            dataset=dataset,
            X=X,
            Y=Y,
            idcs=torch.arange(3),
        )
        self.assertTrue(torch.equal(cv_fold.test_X, X[:3]))
        self.assertTrue(torch.equal(cv_fold.test_Y, Y[:3]))
        self.assertTrue(torch.equal(cv_fold.train_dataset.X, X[3:]))
        self.assertTrue(torch.equal(cv_fold.train_dataset.Y, Y[3:]))
        self.assertTrue(
            torch.equal(cv_fold.train_dataset.Yvar, Yvar[3:])  # pyre-ignore[6]
        )

    def test_model_config(self) -> None:
        # Test that model identifier is correctly computed.
        mc1 = ModelConfig(
            botorch_model_class=SingleTaskGP,
            covar_module_class=ScaleMaternKernel,
            covar_module_options={"ard_num_dims": 1},
            name="GP",
        )
        self.assertEqual(mc1.identifier, "GP")
        mc2 = ModelConfig(
            botorch_model_class=SingleTaskGP,
            covar_module_class=ScaleMaternKernel,
            covar_module_options={"ard_num_dims": 1},
        )
        mc_str = (
            "ModelConfig("
            "botorch_model_class=<class 'botorch.models.gp_regression.SingleTaskGP'>, "
            "model_options={}, mll_class=None, mll_options={}, "
            "input_transform_classes=<class 'botorch.utils.types.DEFAULT'>, "
            "input_transform_options={}, outcome_transform_classes=None, "
            "outcome_transform_options={}, "
            "covar_module_class=<class 'ax.generators.torch.botorch_modular.kernels."
            "ScaleMaternKernel'>, covar_module_options={'ard_num_dims': 1}, "
            "likelihood_class=None, likelihood_options={}, name=None)"
        )
        self.assertEqual(mc2.identifier, mc_str)

    def test_predict_from_model_ensemble_posterior(self) -> None:
        """Test predict_from_model with EnsemblePosterior support."""
        X = torch.rand(2, 3)

        # Create a mock EnsemblePosterior
        mock_posterior = Mock(spec=EnsemblePosterior)
        mock_posterior.mixture_mean = torch.rand(2, 2)
        mock_posterior.mixture_variance = torch.rand(2, 2)

        # Create a mock model
        mock_model = Mock()
        mock_model.posterior.return_value = mock_posterior

        # Test prediction
        mean, cov = predict_from_model(mock_model, X, use_posterior_predictive=False)

        # Verify shapes
        self.assertEqual(mean.shape, (2, 2))  # (n_points, n_outputs)
        self.assertEqual(cov.shape, (2, 2, 2))  # (n_points, n_outputs, n_outputs)
        self.assertTrue(torch.all(cov >= 0))  # Ensure covariance is positive



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.core.auxiliary import AuxiliaryExperiment
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_experiment, get_experiment_with_data


class AuxiliaryExperimentTest(TestCase):
    def test_AuxiliaryExperiment(self) -> None:
        for get_exp_func in [get_experiment, get_experiment_with_data]:
            exp = get_exp_func()
            data = exp.lookup_data()

            # Test init
            aux_exp = AuxiliaryExperiment(experiment=exp)
            self.assertEqual(aux_exp.experiment, exp)
            self.assertEqual(aux_exp.data, data)

            another_aux_exp = AuxiliaryExperiment(
                experiment=exp, data=exp.lookup_data()
            )
            self.assertEqual(another_aux_exp.experiment, exp)
            self.assertEqual(another_aux_exp.data, data)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis, ErrorAnalysisCard
from ax.analysis.analysis_card import AnalysisCardGroup
from ax.analysis.diagnostics import DiagnosticAnalysis
from ax.analysis.healthcheck.can_generate_candidates import (
    CanGenerateCandidatesAnalysis,
)
from ax.analysis.healthcheck.constraints_feasibility import (
    ConstraintsFeasibilityAnalysis,
)
from ax.analysis.healthcheck.healthcheck_analysis import HealthcheckAnalysisCard
from ax.analysis.healthcheck.metric_fetching_errors import MetricFetchingErrorsAnalysis
from ax.analysis.healthcheck.search_space_analysis import SearchSpaceAnalysis
from ax.analysis.healthcheck.should_generate_candidates import ShouldGenerateCandidates
from ax.analysis.insights import InsightsAnalysis
from ax.analysis.results import ResultsAnalysis
from ax.analysis.trials import AllTrialsAnalysis
from ax.core.experiment import Experiment
from ax.core.trial_status import TrialStatus
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import none_throws, override


HEALTH_CHECK_CARDGROUP_TITLE = "Health Checks"
HEALTH_CHECK_CARDGROUP_SUBTITLE = (
    "Comprehensive health checks designed to identify potential issues in the Ax "
    "experiment. These checks cover areas such as metric fetching, search space "
    "configuration, and candidate generation, with the aim of flagging areas where "
    "user intervention may be necessary to ensure the experiment's robustness "
    "and success."
)

OVERVIEW_CARDGROUP_TITLE = "Overview of the Entire Optimization Process "
OVERVIEW_CARDGROUP_SUBTITLE = (
    "This analysis provides an overview of the entire optimization process. "
    "It includes visualizations of the results obtained so far, insights into "
    "the parameter and metric relationships learned by the Ax model, diagnostics "
    "such as model fit, and health checks to assess the overall health of the "
    "experiment."
)


class OverviewAnalysis(Analysis):
    """
    Top-level Analysis that provides an overview of the entire optimization process,
    including results, insights, and diagnostics. OverviewAnalysis examines the
    Experiment and GenerationStrategy's configuration and their respective current
    states to heuristically determine which Analyses to compute under the hood.

    AnalysisCards will be returned in the following groups:
        * Overview
            * Results
                * Pairs of Modeled and Raw ArmEffectsPlots for objectives and
                    constraints
                * Modeled ScatterPlots for objectives versus objectives and objectives
                    versus constraints
                * ParallelCoordinatesPlot for objectives
                * Summary
            * Insights
                * Sensitivity Plots
                * Slice Plots
                * Contour Plots
            * Diagnostic
                * CrossValidationPlots
            * Health Checks
                * MetricFetchingErrorsAnalysis
                * CanGenerateCandidatesAnalysis
                * ConstraintsFeasibilityAnalysis
                * SearchSpaceAnalysis
                * ShouldGenerateCandidates
            * Trial-Level Analyses
                * Trial 0
                    * ArmEffectsPlot
                ...
    """

    def __init__(
        self,
        can_generate: bool | None = None,
        can_generate_reason: str | None = None,
        can_generate_days_till_fail: int | None = None,
        should_generate: bool | None = None,
        should_generate_reason: str | None = None,
    ) -> None:
        super().__init__()
        self.can_generate = can_generate
        self.can_generate_reason = can_generate_reason
        self.can_generate_days_till_fail = can_generate_days_till_fail
        self.should_generate = should_generate
        self.should_generate_reason = should_generate_reason

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardGroup:
        # Compute the arm effects plots, scatter plots, etc.
        results_group = ResultsAnalysis().compute_or_error_card(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        # Compute the sensitivity plots, slice plots, contour plots, etc.
        insights_group = InsightsAnalysis().compute_or_error_card(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        # Compute the diagnostics section (cross validation plots)
        diagnostics_group = DiagnosticAnalysis().compute_or_error_card(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        health_check_analyses = [
            MetricFetchingErrorsAnalysis(),
            CanGenerateCandidatesAnalysis(
                can_generate_candidates=self.can_generate,
                reason=self.can_generate_reason,
                days_till_fail=self.can_generate_days_till_fail,
            )
            if self.can_generate is not None
            and self.can_generate_reason is not None
            and self.can_generate_days_till_fail is not None
            else None,
            ConstraintsFeasibilityAnalysis(),
            *[
                SearchSpaceAnalysis(trial_index=trial.index)
                for trial in none_throws(experiment).trials_by_status[
                    TrialStatus.CANDIDATE
                ]
            ],
            *[
                ShouldGenerateCandidates(
                    should_generate=self.should_generate,
                    reason=self.should_generate_reason,
                    trial_index=trial.index,
                )
                for trial in none_throws(experiment).trials_by_status[
                    TrialStatus.CANDIDATE
                ]
                if self.should_generate is not None
                and self.should_generate_reason is not None
            ],
        ]

        health_check_cards = [
            analyis.compute_or_error_card(
                experiment=experiment,
                generation_strategy=generation_strategy,
                adapter=adapter,
            )
            for analyis in health_check_analyses
            if analyis is not None
        ]

        non_passing_health_checks = [
            card
            for card in health_check_cards
            if (isinstance(card, HealthcheckAnalysisCard) and not card.is_passing())
            or isinstance(card, ErrorAnalysisCard)
        ]

        health_checks_group = (
            AnalysisCardGroup(
                name="HealthchecksAnalysis",
                title=HEALTH_CHECK_CARDGROUP_TITLE,
                subtitle=HEALTH_CHECK_CARDGROUP_SUBTITLE,
                children=non_passing_health_checks,
            )
            if len(non_passing_health_checks) > 0
            else None
        )

        trials_group = AllTrialsAnalysis().compute_or_error_card(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        return self._create_analysis_card_group(
            title=OVERVIEW_CARDGROUP_TITLE,
            subtitle=OVERVIEW_CARDGROUP_SUBTITLE,
            children=[
                group
                for group in [
                    results_group,
                    insights_group,
                    diagnostics_group,
                    health_checks_group,
                    trials_group,
                ]
                if group is not None
            ],
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from itertools import chain

from typing import Any

import torch
from ax.core.search_space import SearchSpaceDigest
from ax.utils.common.typeutils import _argparse_type_encoder
from botorch.models.transforms.input import (
    FilterFeatures,
    InputPerturbation,
    InputTransform,
    Normalize,
    Warp,
)
from botorch.utils.datasets import MultiTaskDataset, RankingDataset, SupervisedDataset
from botorch.utils.dispatcher import Dispatcher
from botorch.utils.transforms import normalize_indices
from pyre_extensions import none_throws


input_transform_argparse = Dispatcher(
    name="input_transform_argparse", encoder=_argparse_type_encoder
)


def _set_default_bounds(
    search_space_digest: SearchSpaceDigest,
    input_transform_options: dict[str, Any],
    d: int,
    torch_device: torch.device | None = None,
    torch_dtype: torch.dtype | None = None,
) -> None:
    """Set default bounds in input_transform_options, in-place.

    Args:
        search_space_digest: Search space digest.
        input_transform_options: Input transform kwargs.
        d: The dimension of the input space.
        torch_device: The device on which the input transform will be used.
        torch_dtype: The dtype on which the input transform will be used.
    """
    bounds = torch.as_tensor(
        search_space_digest.bounds,
        dtype=torch_dtype,
        device=torch_device,
    ).T

    if (
        ("bounds" not in input_transform_options)
        and (bounds.shape[-1] < d)
        and (len(search_space_digest.task_features) == 0)
    ):
        raise NotImplementedError(
            "Normalization bounds should be specified explicitly if there"
            " are task features outside the search space."
        )

    input_transform_options.setdefault("bounds", bounds)


@input_transform_argparse.register(InputTransform)
def _input_transform_argparse_base(
    input_transform_class: type[InputTransform],
    dataset: SupervisedDataset | None = None,
    search_space_digest: SearchSpaceDigest | None = None,
    input_transform_options: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """
    Extract the input transform kwargs from the given arguments.

    Args:
        input_transform_class: Input transform class.
        dataset: Dataset containing feature matrix and the response.
        search_space_digest: Search space digest.
        input_transform_options: An optional dictionary of input transform options.
            This may include overrides for the above options. For example, when
            `input_transform_class` is Normalize this dictionary might include
            {
                "d": 2, # the dimension of the input space
            }
            See `botorch.models.transforms.input.py` for more options.

    Returns:
        A dictionary with input transform kwargs.
    """
    return input_transform_options or {}


@input_transform_argparse.register(Warp)
def _input_transform_argparse_warp(
    input_transform_class: type[Warp],
    dataset: SupervisedDataset,
    search_space_digest: SearchSpaceDigest,
    input_transform_options: dict[str, Any] | None = None,
    torch_device: torch.device | None = None,
    torch_dtype: torch.dtype | None = None,
) -> dict[str, Any]:
    """Extract the base input transform kwargs form the given arguments.

    Args:
        input_transform_class: Input transform class.
        dataset: Dataset containing feature matrix and the response.
        search_space_digest: Search space digest.
        input_transform_options: Input transform kwargs.
        torch_device: The device on which the input transform will be used.
        torch_dtype: The dtype on which the input transform will be used.

    Returns:
        A dictionary with input transform kwargs.
    """
    input_transform_options = input_transform_options or {}
    d = len(dataset.feature_names)
    indices = list(range(d))
    task_features = none_throws(
        normalize_indices(search_space_digest.task_features, d=d)
    )

    for task_feature in sorted(task_features, reverse=True):
        del indices[task_feature]

    input_transform_options.setdefault("d", d)
    input_transform_options.setdefault("indices", indices)

    # if ranking dataset, infer the bounds instead as it may be unbounded
    if not isinstance(dataset, RankingDataset):
        _set_default_bounds(
            search_space_digest=search_space_digest,
            input_transform_options=input_transform_options,
            d=d,
            torch_device=torch_device,
            torch_dtype=torch_dtype,
        )
    return input_transform_options


@input_transform_argparse.register(Normalize)
def _input_transform_argparse_normalize(
    input_transform_class: type[Normalize],
    dataset: SupervisedDataset,
    search_space_digest: SearchSpaceDigest,
    input_transform_options: dict[str, Any] | None = None,
    torch_device: torch.device | None = None,
    torch_dtype: torch.dtype | None = None,
) -> dict[str, Any]:
    """
    Extract the base input transform kwargs form the given arguments.
    NOTE: This input constructor doesn't support the case when there are
    task features that are not part of search space digest. In that case,
    the bounds need to be specified manually as a part of input_transform_options.

    Args:
        input_transform_class: Input transform class.
        dataset: Dataset containing feature matrix and the response.
        search_space_digest: Search space digest.
        input_transform_options: Input transform kwargs.
        torch_device: The device on which the input transform will be used.
        torch_dtype: The dtype on which the input transform will be used.

    Returns:
        A dictionary with input transform kwargs.
    """
    input_transform_options = input_transform_options or {}
    if isinstance(dataset, MultiTaskDataset) and dataset.has_heterogeneous_features:
        # set d to number of features in the full feature space
        d = len(set(chain(*(ds.feature_names for ds in dataset.datasets.values()))))
    else:
        d = input_transform_options.get("d", len(dataset.feature_names))
    input_transform_options["d"] = d

    if isinstance(dataset, RankingDataset):
        input_transform_options["indices"] = None
    elif input_transform_options.get("indices") is None:
        indices = list(range(d))
        if isinstance(dataset, MultiTaskDataset) and dataset.has_heterogeneous_features:
            # task feature is last feature
            del indices[none_throws(dataset.task_feature_index)]
        else:
            task_features = none_throws(
                normalize_indices(search_space_digest.task_features, d=d)
            )
            for task_feature in sorted(task_features, reverse=True):
                del indices[task_feature]

        if len(indices) < d:
            input_transform_options["indices"] = indices

    # if ranking dataset, infer the bounds instead as it may be unbounded
    if not isinstance(dataset, RankingDataset):
        _set_default_bounds(
            search_space_digest=search_space_digest,
            input_transform_options=input_transform_options,
            d=d,
            torch_device=torch_device,
            torch_dtype=torch_dtype,
        )

    return input_transform_options


@input_transform_argparse.register(FilterFeatures)
def _input_transform_argparse_filter_features(
    input_transform_class: type[FilterFeatures],
    dataset: SupervisedDataset,
    search_space_digest: SearchSpaceDigest,
    input_transform_options: dict[str, Any] | None = None,
    torch_device: torch.device | None = None,
    torch_dtype: torch.dtype | None = None,
) -> dict[str, Any]:
    """Extract the FilterFeatures input transform kwargs from the given arguments.

    Args:
        input_transform_class: Input transform class.
        dataset: Dataset containing feature matrix and the response.
        search_space_digest: Search space digest.
        input_transform_options: Input transform kwargs. May contain:
            - "feature_indices": Explicit list of feature indices to keep
            - "ignored_params": List of parameter names to ignore
            - Other FilterFeatures kwargs
        torch_device: The device on which the input transform will be used.
        torch_dtype: The dtype on which the input transform will be used.

    Returns:
        A dictionary with FilterFeatures kwargs.
    """
    input_transform_options_copy = (
        input_transform_options.copy() if input_transform_options else {}
    )

    # If no options are provided, keep all features
    if not input_transform_options_copy:
        return {"feature_indices": torch.arange(len(dataset.feature_names))}

    feature_names = dataset.feature_names

    # Validate ignored_params if present
    if "ignored_params" in input_transform_options_copy:
        ignored_params = input_transform_options_copy["ignored_params"]
        invalid_params = [
            param for param in ignored_params if param not in feature_names
        ]
        # TO DO: This may error out on Categorical parameters that went through
        # `OneHot` transform. We should add the ability to handle this in the future.
        if invalid_params:
            raise ValueError(
                f"Invalid parameter names in ignored_params: {invalid_params}. "
                f"Valid feature names are: {feature_names}."
            )

    # If feature_indices is already provided, use it directly
    if "feature_indices" in input_transform_options_copy:
        if "ignored_params" in input_transform_options_copy:
            # If both feature_indices and ignored_params are provided, check for
            # consistency and pop "ignored_params"
            feature_indices = input_transform_options_copy["feature_indices"]
            filtered_indices_from_ignored_param = torch.tensor(
                [
                    i
                    for i, name in enumerate(feature_names)
                    if name not in input_transform_options_copy["ignored_params"]
                ],
                dtype=torch.int64,
            )

            if not torch.equal(feature_indices, filtered_indices_from_ignored_param):
                raise ValueError(
                    f"Filtered features passed in by feature_indices {feature_indices} "
                    "is inconsistent with filtered feature indices computed from "
                    f"ignored_params {filtered_indices_from_ignored_param}."
                    "Please provide only one of 'feature_indices' or 'ignored_params', "
                    "or ensure they are consistent."
                )
            # pop "ignored_params" as it is not an expected arg of FilterFeatures
            input_transform_options_copy.pop("ignored_params")
        return input_transform_options_copy

    # If only ignored_params is provided, compute feature_indices from it
    # and find feature_indices to keep
    if "ignored_params" in input_transform_options_copy:
        ignored_params = input_transform_options_copy.pop("ignored_params")

        feature_indices = [
            i for i, name in enumerate(feature_names) if name not in ignored_params
        ]

        input_transform_options_copy["feature_indices"] = torch.tensor(
            feature_indices, dtype=torch.int64
        )

    return input_transform_options_copy


@input_transform_argparse.register(InputPerturbation)
def _input_transform_argparse_input_perturbation(
    input_transform_class: type[InputPerturbation],
    search_space_digest: SearchSpaceDigest,
    dataset: SupervisedDataset | None = None,
    input_transform_options: dict[str, Any] | None = None,
    torch_device: torch.device | None = None,
    torch_dtype: torch.dtype | None = None,
) -> dict[str, Any]:
    """Extract the base input transform kwargs form the given arguments.

    Args:
        input_transform_class: Input transform class.
        dataset: Dataset containing feature matrix and the response.
        search_space_digest: Search space digest.
        input_transform_options: Input transform kwargs.
        torch_device: The device on which the input transform will be used.
        torch_dtype: The dtype on which the input transform will be used.

    Returns:
        A dictionary with input transform kwargs.
    """

    input_transform_options = input_transform_options or {}

    robust_digest = search_space_digest.robust_digest

    if robust_digest is None:
        raise ValueError("Robust search space digest must be provided.")

    if len(robust_digest.environmental_variables) > 0:
        # TODO[T131759269]: support env variables.
        raise NotImplementedError(
            "Environmental variable support is not yet implemented."
        )
    if robust_digest.sample_param_perturbations is None:
        raise ValueError("Robust digest needs to sample parameter perturbations.")

    samples = torch.as_tensor(
        robust_digest.sample_param_perturbations(),
        dtype=torch_dtype,
        device=torch_device,
    )

    input_transform_options.setdefault("perturbation_set", samples)

    input_transform_options.setdefault("multiplicative", robust_digest.multiplicative)

    return input_transform_options



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import logging

from logging import Logger
from typing import Any

import numpy as np

import pandas as pd
from ax.core.base_trial import BaseTrial
from ax.core.map_data import MAP_KEY, MapData
from ax.core.map_metric import MapMetric
from ax.core.metric import Metric, MetricFetchE, MetricFetchResult
from ax.core.trial import Trial
from ax.utils.common.logger import get_logger
from ax.utils.common.result import Err, Ok
from pyre_extensions import assert_is_instance

logger: Logger = get_logger(__name__)

# Default in Tensorboard UI (https://fburl.com/workplace/1sq11640)
SMOOTHING_DEFAULT = 0
RUN_METADATA_KEY = "tb_log_dir"

try:
    from tensorboard.backend.event_processing import (
        plugin_event_multiplexer as event_multiplexer,
    )

    logging.getLogger("tensorboard").setLevel(logging.CRITICAL)

    class TensorboardMetric(MapMetric):
        """A *new* `MapMetric` for getting Tensorboard metrics."""

        def __init__(
            self,
            name: str,
            tag: str,
            lower_is_better: bool | None = True,
            smoothing: float = SMOOTHING_DEFAULT,
            cumulative_best: bool = False,
            percentile: float | None = None,
        ) -> None:
            """
            Args:
                name: The name of the metric.
                tag: The name of the learning curve in the Tensorboard Scalars tab.
                lower_is_better: If True, lower curve values are considered better.
                smoothing: If > 0, apply exponential weighted mean to the curve. This
                    is the same postprocessing as the "smoothing" slider in the
                    Tensorboard UI.
                cumulative_best: If True, for each trial, apply cumulative best to
                    the curve (i.e., if lower is better, then we return a curve
                    representing the cumulative min of the raw curve).
                percentile: If not None, return the (rolling) percentile value
                    of the curve.
                    e.g. if the original curve is [0, 6, 4, 2] and percentile=0.5, then
                    the returned curve is [0, 3, 4, 3]. Rolling percentile is applied
                    after any potential smoothing or cumulative_best processing.
            """
            super().__init__(name=name, lower_is_better=lower_is_better)

            self.smoothing = smoothing
            self.tag = tag
            self.cumulative_best = cumulative_best
            self.percentile = percentile

        @classmethod
        def is_available_while_running(cls) -> bool:
            return True

        def bulk_fetch_trial_data(
            self, trial: BaseTrial, metrics: list[Metric], **kwargs: Any
        ) -> dict[str, MetricFetchResult]:
            """Fetch multiple metrics data for one trial, using instance attributes
            of the metrics.

            Returns Dict of metric_name => Result
            Default behavior calls `fetch_trial_data` for each metric. Subclasses should
            override this to perform trial data computation for multiple metrics.
            """
            tb_metrics = [
                assert_is_instance(metric, TensorboardMetric) for metric in metrics
            ]

            trial = assert_is_instance(trial, Trial)
            if trial.arm is None:
                raise ValueError("Trial must have arm set.")

            arm_name = trial.arm.name

            try:
                mul = self._get_event_multiplexer_for_trial(trial=trial)
            except Exception as e:
                return {
                    metric.name: Err(
                        MetricFetchE(
                            message=f"Failed to get event multiplexer for {trial=}",
                            exception=e,
                        )
                    )
                    for metric in tb_metrics
                }

            scalar_dict = mul.PluginRunToTagToContent("scalars")
            if len(scalar_dict) == 0:
                return {
                    metric.name: Err(
                        MetricFetchE(
                            message=(
                                "Tensorboard multiplexer is empty. This can happen if "
                                "TB data is not populated at the time of fetch. Check "
                                "the corresponding logs to confirm that Tensorboard "
                                "data is available."
                            ),
                            exception=None,
                        )
                    )
                    for metric in tb_metrics
                }

            res = {}
            for metric in tb_metrics:
                try:
                    records = [
                        {
                            "trial_index": trial.index,
                            "arm_name": arm_name,
                            "metric_name": metric.name,
                            MAP_KEY: t.step,
                            "mean": (
                                t.tensor_proto.double_val[0]
                                if t.tensor_proto.double_val
                                else t.tensor_proto.float_val[0]
                            ),
                            "sem": float("nan"),
                        }
                        for run_name, tags in scalar_dict.items()
                        for tag in tags
                        if tag == metric.tag
                        for t in mul.Tensors(run_name, tag)
                    ]

                    # If records is empty something has gone wrong: either the tag is
                    # not present on the multiplexer or the content referenced is empty
                    if len(records) == 0:
                        if metric.tag not in [
                            j for sub in scalar_dict.values() for j in sub
                        ]:
                            raise KeyError(
                                f"Tag {metric.tag} not found on multiplexer {mul=}. "
                                "Did you specify this tag exactly as it appears in "
                                "the TensorBoard UI's Scalars tab?"
                            )
                        else:
                            raise ValueError(
                                f"Found tag {metric.tag}, but no data found for it. Is "
                                "the curve empty in the TensorBoard UI?"
                            )
                    df = self._process_records_to_df(metric=metric, records=records)
                    # Accumulate successfully extracted timeseries
                    res[metric.name] = Ok(
                        MapData(
                            df=df,
                        )
                    )

                except Exception as e:
                    res[metric.name] = Err(
                        MetricFetchE(
                            message=f"Failed to fetch data for {metric.name}",
                            exception=e,
                        )
                    )

            self._clear_multiplexer_if_possible(multiplexer=mul)

            return res

        def fetch_trial_data(
            self, trial: BaseTrial, **kwargs: Any
        ) -> MetricFetchResult:
            """Fetch data for one trial."""

            return self.bulk_fetch_trial_data(trial=trial, metrics=[self], **kwargs)[
                self.name
            ]

        def _get_event_multiplexer_for_trial(
            self, trial: BaseTrial
        ) -> event_multiplexer.EventMultiplexer:
            """Get an event multiplexer with the logs for a given trial."""

            mul = event_multiplexer.EventMultiplexer(max_reload_threads=20)
            mul.AddRunsFromDirectory(trial.run_metadata[RUN_METADATA_KEY], None)
            mul.Reload()

            return mul

        def _clear_multiplexer_if_possible(
            self, multiplexer: event_multiplexer.EventMultiplexer
        ) -> None:
            """
            Clear the multiplexer of all data. This is a no-op here, but for some
            Multiplexers which may implement a clearing method this method can be
            important for managing memory consumption.
            """
            pass

        def _process_records_to_df(
            self, metric: TensorboardMetric, records: list[dict[str, Any]]
        ) -> pd.DataFrame:
            """Process records to a MapData dataframe."""
            df = (
                pd.DataFrame(records)
                # If a metric has multiple records for the same arm, metric, and
                # step (sometimes caused by restarts, etc) take the mean
                .groupby(["arm_name", "metric_name", MAP_KEY])
                .mean()
                .reset_index()
            )

            # If all values are NaNs or Infs, we raise an error
            # If some values are NaNs or Infs, we log a warning and filter out the
            # non-finite values
            is_finite = np.isfinite(df["mean"])
            is_not_finite = ~is_finite
            if np.all(is_not_finite):
                raise ValueError("All values are NaNs or Infs.")

            if np.any(is_not_finite):
                logger.warning(
                    f"{sum(is_not_finite)} / {len(is_not_finite)} data points are NaNs "
                    "or Infs. Filtering out non-finite values."
                )
                df = df[is_finite]

            # Apply per-metric post-processing
            # Apply cumulative "best" (min if lower_is_better)
            if metric.cumulative_best:
                if metric.lower_is_better:
                    df["mean"] = df["mean"].cummin()
                else:
                    df["mean"] = df["mean"].cummax()

            # Apply smoothing
            if metric.smoothing > 0:
                df["mean"] = df["mean"].ewm(alpha=1 - metric.smoothing).mean()

            # Apply rolling percentile
            if metric.percentile is not None:
                df["mean"] = df["mean"].expanding().quantile(metric.percentile)

            return df

except ImportError:
    logger.warning(
        "tensorboard package not found. If you would like to use "
        "TensorboardMetric, please install tensorboard."
    )
    pass



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy

from ax.adapter.base import DataLoaderConfig
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.transforms.fixed_to_tunable import FixedToTunable
from ax.core.observation import ObservationFeatures
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.parameter_constraint import OrderConstraint
from ax.core.search_space import SearchSpace
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_experiment_with_observations
from pandas.testing import assert_frame_equal
from pyre_extensions import assert_is_instance


class FixedToTunableTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        # Search space with fixed parameters that should be converted to tunable
        self.search_space_with_fixed = SearchSpace(
            parameters=[
                RangeParameter(
                    "a", lower=1, upper=3, parameter_type=ParameterType.FLOAT
                ),
                FixedParameter("b", parameter_type=ParameterType.FLOAT, value=2.0),
                ChoiceParameter(
                    "c", parameter_type=ParameterType.STRING, values=["x", "y", "z"]
                ),
                FixedParameter("d", parameter_type=ParameterType.INT, value=5),
                FixedParameter("e", parameter_type=ParameterType.FLOAT, value=3.0),
            ]
        )

        # Joint search space with range parameters for the fixed parameters
        self.joint_search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "a", lower=1, upper=3, parameter_type=ParameterType.FLOAT
                ),
                RangeParameter(
                    "b", lower=1, upper=5, parameter_type=ParameterType.FLOAT
                ),
                ChoiceParameter(
                    "c", parameter_type=ParameterType.STRING, values=["x", "y", "z"]
                ),
                RangeParameter(
                    "d", lower=1, upper=10, parameter_type=ParameterType.INT
                ),
                FixedParameter("e", parameter_type=ParameterType.FLOAT, value=4.0),
            ]
        )

        self.t = FixedToTunable(search_space=self.joint_search_space)

    def test_init(self) -> None:
        self.assertEqual(self.t.search_space, self.joint_search_space)

    def test_init_requires_search_space(self) -> None:
        with self.assertRaisesRegex(
            AssertionError, "FixedToTunable requires search space"
        ):
            FixedToTunable(search_space=None)

    def test_transform_search_space(self) -> None:
        # Test transforming fixed parameters to tunable ones
        transformed_ss = self.t.transform_search_space(self.search_space_with_fixed)

        # Parameter 'a' should remain unchanged (already a RangeParameter)
        self.assertEqual(
            transformed_ss.parameters["a"], self.search_space_with_fixed.parameters["a"]
        )

        # Parameter 'b' should be converted from FixedParameter to RangeParameter
        transformed_b = assert_is_instance(
            transformed_ss.parameters["b"], RangeParameter
        )
        self.assertEqual(transformed_b.lower, 1)
        self.assertEqual(transformed_b.upper, 5)
        self.assertEqual(transformed_b.parameter_type, ParameterType.FLOAT)

        # Parameter 'c' should remain unchanged (ChoiceParameter)
        self.assertEqual(
            transformed_ss.parameters["c"], self.search_space_with_fixed.parameters["c"]
        )

        # Parameter 'd' should be converted from FixedParameter to RangeParameter
        transformed_d = assert_is_instance(
            transformed_ss.parameters["d"], RangeParameter
        )
        self.assertEqual(transformed_d.lower, 1)
        self.assertEqual(transformed_d.upper, 10)
        self.assertEqual(transformed_d.parameter_type, ParameterType.INT)

        # Parameter 'e' should remain as FixedParameter (with the original value from
        # self.search_space_with_fixed) since it's fixed in the joint search space
        transformed_e = assert_is_instance(
            transformed_ss.parameters["e"], FixedParameter
        )
        self.assertEqual(transformed_e.value, 3.0)

    def test_transform_search_space_with_constraints(self) -> None:
        # Create search space with parameter constraints
        parameters = [
            RangeParameter("x", lower=1, upper=3, parameter_type=ParameterType.FLOAT),
            RangeParameter("y", lower=1, upper=3, parameter_type=ParameterType.FLOAT),
            FixedParameter("z", parameter_type=ParameterType.FLOAT, value=2.0),
        ]
        search_space_with_constraints = SearchSpace(
            parameters=parameters,
            parameter_constraints=[
                OrderConstraint(
                    lower_parameter=parameters[0], upper_parameter=parameters[1]
                )
            ],
        )

        # Joint space with range parameter for 'y'
        joint_space_with_y_range = SearchSpace(
            parameters=[
                RangeParameter(
                    "x", lower=1, upper=3, parameter_type=ParameterType.FLOAT
                ),
                RangeParameter(
                    "y", lower=1, upper=5, parameter_type=ParameterType.FLOAT
                ),
                RangeParameter(
                    "z", lower=1, upper=5, parameter_type=ParameterType.FLOAT
                ),
            ]
        )

        t = FixedToTunable(search_space=joint_space_with_y_range)
        transformed_ss = t.transform_search_space(search_space_with_constraints)

        # Verify parameter 'z' is transformed to RangeParameter
        transformed_z = assert_is_instance(
            transformed_ss.parameters["z"], RangeParameter
        )
        self.assertEqual(transformed_z.lower, 1)
        self.assertEqual(transformed_z.upper, 5)

        # Verify constraints are preserved
        self.assertEqual(len(transformed_ss.parameter_constraints), 1)
        self.assertIs(
            transformed_ss.parameter_constraints[0],
            search_space_with_constraints.parameter_constraints[0],
        )

    def test_transform_experiment_data(self) -> None:
        # Test that experiment data is returned unchanged
        parameterizations = [
            {"a": 1.0, "b": 2.0, "c": "x", "d": 5, "e": 3.0},
            {"a": 2.0, "b": 2.0, "c": "y", "d": 5, "e": 3.0},
            {"a": 3.0, "b": 2.0, "c": "z", "d": 5, "e": 3.0},
        ]
        experiment = get_experiment_with_observations(
            observations=[[1.0], [2.0], [3.0]],
            search_space=self.search_space_with_fixed,
            parameterizations=parameterizations,
        )
        experiment_data = extract_experiment_data(
            experiment=experiment, data_loader_config=DataLoaderConfig()
        )

        original_data = deepcopy(experiment_data)
        transformed_data = self.t.transform_experiment_data(experiment_data)

        # The transform should return the data unchanged
        self.assertIs(transformed_data, experiment_data)

        # Verify all data is identical
        assert_frame_equal(transformed_data.arm_data, original_data.arm_data)
        assert_frame_equal(
            transformed_data.observation_data, original_data.observation_data
        )

    def test_transform_observation_features(self) -> None:
        # Since FixedToTunable doesn't override transform_observation_features,
        # it should use the base class implementation which returns unchanged data
        observation_features = [
            ObservationFeatures(
                parameters={"a": 2.0, "b": 2.0, "c": "y", "d": 5, "e": 3.0}
            )
        ]

        # Test transform (should be unchanged)
        transformed_obs = self.t.transform_observation_features(
            deepcopy(observation_features)
        )
        self.assertEqual(transformed_obs, observation_features)

        # Test untransform (should also be unchanged)
        untransformed_obs = self.t.untransform_observation_features(
            deepcopy(observation_features)
        )
        self.assertEqual(untransformed_obs, observation_features)

    def test_clone_behavior(self) -> None:
        # Test that transformed parameters are properly cloned
        transformed_ss = self.t.transform_search_space(self.search_space_with_fixed)
        original_b_param = assert_is_instance(
            self.joint_search_space.parameters["b"], RangeParameter
        )
        transformed_b_param = assert_is_instance(
            transformed_ss.parameters["b"], RangeParameter
        )

        # They should have the same values but be different objects
        self.assertEqual(original_b_param.lower, transformed_b_param.lower)
        self.assertEqual(
            original_b_param.upper,
            transformed_b_param.upper,
        )
        self.assertEqual(
            original_b_param.parameter_type, transformed_b_param.parameter_type
        )
        self.assertIsNot(original_b_param, transformed_b_param)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Mapping

from ax.core.arm import Arm
from ax.core.types import TParamValue


def int_from_params(
    params: Mapping[str, TParamValue], n_possibilities: int = 10
) -> int:
    """
    Get an int between 0 and n_possibilities - 1, using a hash of the parameters.
    """
    arm_hash = Arm.md5hash(parameters=params)
    return int(arm_hash[-1], base=16) % n_possibilities



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import time
from collections import OrderedDict
from collections.abc import Mapping, MutableMapping, Sequence
from copy import deepcopy
from dataclasses import dataclass, field
from logging import Logger
from typing import Any

import numpy as np
import pandas as pd
from ax.adapter.data_utils import (
    DataLoaderConfig,
    ExperimentData,
    extract_experiment_data,
)
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.cast import Cast
from ax.adapter.transforms.fill_missing_parameters import FillMissingParameters
from ax.core.arm import Arm
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import extract_arm_predictions, GeneratorRun
from ax.core.map_data import MAP_KEY
from ax.core.observation import (
    Observation,
    ObservationData,
    ObservationFeatures,
    recombine_observations,
)
from ax.core.optimization_config import OptimizationConfig
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.core.types import (
    TCandidateMetadata,
    TModelCov,
    TModelMean,
    TModelPredict,
    TParameterization,
)
from ax.core.utils import get_target_trial_index, has_map_metrics
from ax.exceptions.core import UnsupportedError, UserInputError
from ax.exceptions.model import AdapterMethodNotImplementedError, ModelError
from ax.generators.base import Generator
from ax.generators.types import TConfig
from ax.utils.common.logger import get_logger
from botorch.settings import validate_input_scaling
from pandas import DataFrame
from pyre_extensions import assert_is_instance, none_throws

logger: Logger = get_logger(__name__)


@dataclass(frozen=True)
class BaseGenArgs:
    search_space: SearchSpace
    optimization_config: OptimizationConfig | None
    pending_observations: dict[str, list[ObservationFeatures]]
    fixed_features: ObservationFeatures


@dataclass(frozen=True)
class GenResults:
    observation_features: list[ObservationFeatures]
    weights: list[float]
    best_observation_features: ObservationFeatures | None = None
    gen_metadata: dict[str, Any] = field(default_factory=dict)


class Adapter:
    """The main object for using generators in Ax.

    Adapter specifies 3 methods for using generators:

    - predict: Make predictions with the generator. This method is not optimized for
      speed and so should be used primarily for plotting or similar tasks
      and not inside an optimization loop.
    - gen: Use the generator to generate new candidates.
    - cross_validate: Do cross validation to assess generator predictions.

    Adapter converts Ax types like Data and Arm to types that are
    meant to be consumed by the generators. The data sent to the generator will depend
    on the implementation of the subclass, which will specify the actual API
    for external generator.

    This class also applies a sequence of transforms to the input data and
    problem specification which can be used to ensure that the external generator
    receives appropriate inputs.

    Subclasses will implement what is here referred to as the "terminal
    transform", which is a transform that changes types of the data and problem
    specification.
    """

    def __init__(
        self,
        *,
        experiment: Experiment,
        generator: Generator,
        search_space: SearchSpace | None = None,
        data: Data | None = None,
        transforms: Sequence[type[Transform]] | None = None,
        transform_configs: Mapping[str, TConfig] | None = None,
        optimization_config: OptimizationConfig | None = None,
        expand_model_space: bool = True,
        fit_tracking_metrics: bool = True,
        fit_on_init: bool = True,
        data_loader_config: DataLoaderConfig | None = None,
        # fit_out_of_design, fit_abandoned, and fit_only_completed_map_metrics
        # were deprecated in Ax 1.0.0, so they can now be reaped.
        fit_out_of_design: bool | None = None,
        fit_abandoned: bool | None = None,
        fit_only_completed_map_metrics: bool | None = None,
    ) -> None:
        """
        Applies transforms and fits the generator.

        Args:
            experiment: An ``Experiment`` object representing the setup and the
                current state of the experiment, including the search space,
                trials and observation data. It is used to extract various
                attributes, and is not mutated.
            generator: A ``Generator`` that is used for generating candidates.
                Its interface will be specified in subclasses. If generator requires
                initialization, that should be done prior to its use here.
            search_space: An optional ``SearchSpace`` for fitting the generator.
                If not provided, `experiment.search_space` is used.
                The search space may be modified during ``Adapter.gen``, e.g.,
                to try out a different set of parameter bounds or constraints.
                The bounds of the ``RangeParameter``s are considered soft and
                will be expanded to match the range of the data sent in for fitting,
                if `expand_model_space` is True.
            data: An optional ``Data`` object, containing mean and SEM observations.
                If `None`, extracted using `experiment.lookup_data()`.
            transforms: List of uninitialized transform classes. Forward
                transforms will be applied in this order, and untransforms in
                the reverse order.
            transform_configs: A dictionary from transform name to the
                transform config dictionary.
            optimization_config: An optional ``OptimizationConfig`` defining how to
                optimize the generator. Defaults to `experiment.optimization_config`.
            expand_model_space: If True, expand range parameter bounds in model
                space to cover given training data. This will make the modeling
                space larger than the search space if training data fall outside
                the search space. Will also include training points that violate
                parameter constraints in the modeling.
            fit_tracking_metrics: Whether to fit a surrogate model for tracking metrics.
                Setting this to False will improve runtime at the expense of
                models not being available for predicting tracking metrics.
                NOTE: This can only be set to False when the optimization config
                is provided.
            fit_on_init: Whether to fit the generator on initialization. This can
                be used to skip generator fitting when a fitted generator is not needed.
                To fit the generator afterwards, use `_process_and_transform_data`
                to get the transformed inputs and call `_fit_if_implemented` with
                the transformed inputs.
            data_loader_config: A DataLoaderConfig of options for loading data. See the
                docstring of DataLoaderConfig for more details.
            fit_out_of_design: Deprecation warning: `fit_out_of_design` is deprecated.
                Overwrites `data_loader_config.fit_out_of_design` if not None.
            fit_abandoned: Deprecation warning: `fit_out_of_design` is deprecated.
                Overwrites `data_loader_config.fit_abandoned` if not None.
            fit_only_completed_map_metrics: Deprecation warning: `fit_out_of_design`
                is deprecated. If not None, overwrites
                `data_loader_config.fit_only_completed_map_metrics`.
        """
        if data_loader_config is None:
            data_loader_config = DataLoaderConfig()
        self._data_loader_config: DataLoaderConfig = (
            _legacy_overwrite_data_loader_config(
                data_loader_config=data_loader_config,
                fit_out_of_design=fit_out_of_design,
                fit_abandoned=fit_abandoned,
                fit_only_completed_map_metrics=fit_only_completed_map_metrics,
            )
        )

        t_fit_start = time.monotonic()
        transforms = transforms or []
        transforms = [Cast] + list(transforms)
        transform_configs = {} if transform_configs is None else transform_configs
        if "FillMissingParameters" in transform_configs:
            transforms = [FillMissingParameters] + transforms
        self._raw_transforms = transforms
        self._transform_configs: Mapping[str, TConfig] = transform_configs

        self.fit_time: float = 0.0
        self.fit_time_since_gen: float = 0.0
        self._metric_names: set[str] = set()
        # pyre-ignore [13] Assigned in _set_and_filter_training_data.
        self._training_data: ExperimentData
        self._optimization_config: OptimizationConfig | None = optimization_config
        self._training_in_design_idx: list[bool] = []
        self._status_quo: Observation | None = None
        self._status_quo_name: str | None = None
        self.transforms: MutableMapping[str, Transform] = OrderedDict()
        self._model_key: str | None = None
        self._model_kwargs: dict[str, Any] | None = None
        self._bridge_kwargs: dict[str, Any] | None = None
        # The space used for optimization.
        search_space = search_space or experiment.search_space
        self._search_space: SearchSpace = search_space.clone()
        # The space used for modeling. Might be larger than the optimization
        # space to cover training data.
        self._model_space: SearchSpace = search_space.clone()
        self._fit_tracking_metrics = fit_tracking_metrics
        self.outcomes: list[str] = []
        self._experiment_has_immutable_search_space_and_opt_config: bool = (
            experiment is not None and experiment.immutable_search_space_and_opt_config
        )
        self._experiment_properties: dict[str, Any] = experiment._properties
        self._experiment: Experiment = experiment

        if self._optimization_config is None:
            self._optimization_config = experiment.optimization_config
        self._arms_by_signature: dict[str, Arm] = experiment.arms_by_signature

        if self._fit_tracking_metrics is False:
            if self._optimization_config is None:
                raise UserInputError(
                    "Optimization config is required when "
                    "`fit_tracking_metrics` is False."
                )
            self.outcomes = sorted(self._optimization_config.metrics.keys())

        # Set training data (in the raw / untransformed space). This also omits
        # out-of-design and abandoned observations depending on the corresponding flags.
        experiment_data = extract_experiment_data(
            experiment=experiment,
            data_loader_config=self._data_loader_config,
            data=data,
        )
        if expand_model_space:
            self._set_model_space(arm_data=experiment_data.arm_data)
        experiment_data = self._set_and_filter_training_data(
            experiment_data=experiment_data, search_space=self._model_space
        )

        # Set model status quo.
        # NOTE: training data must be set before setting the status quo.
        self._set_status_quo(experiment=experiment)

        # Save generator, apply terminal transform, and fit.
        self.generator = generator
        if fit_on_init:
            experiment_data, search_space = self._transform_data(
                experiment_data=experiment_data,
                search_space=self._model_space,
                transforms=self._raw_transforms,
                transform_configs=self._transform_configs,
            )
            self._fit_if_implemented(
                search_space=search_space,
                experiment_data=experiment_data,
                time_so_far=time.monotonic() - t_fit_start,
            )

    @property
    def can_predict(self) -> bool:
        """Whether this adapter can predict outcomes for new parameterizations."""
        return self.generator.can_predict

    @property
    def can_model_in_sample(self) -> bool:
        """Whether this adapter can model (e.g. apply shrinkage) on observed
        parameterizations (in this case, it needs to support calling `predict`()
        on points in the training data / provided during `fit()`)."""
        return self.generator.can_model_in_sample

    def _fit_if_implemented(
        self,
        search_space: SearchSpace,
        experiment_data: ExperimentData,
        time_so_far: float,
    ) -> None:
        r"""Fits the generator if `_fit` is implemented and stores fit time.

        Args:
            search_space: A transformed search space for fitting the generator.
            experiment_data: The ``ExperimentData`` to fit the generator on, with
                the transforms already applied.
            time_so_far: Time spent in initializing the generator up to
                `_fit_if_implemented` call.
        """
        try:
            t_fit_start = time.monotonic()
            self._fit(search_space=search_space, experiment_data=experiment_data)
            increment = time.monotonic() - t_fit_start + time_so_far
            self.fit_time += increment
            self.fit_time_since_gen += increment
        except AdapterMethodNotImplementedError:
            pass

    def _process_and_transform_data(
        self, experiment: Experiment, data: Data | None = None
    ) -> tuple[ExperimentData, SearchSpace]:
        r"""Processes the data into ``ExperimentData`` and returns the transformed
        ``ExperimentData`` and the search space. This packages the following methods:
        * self._set_and_filter_training_data
        * self._set_status_quo
        * self._transform_data
        """
        experiment_data = extract_experiment_data(
            experiment=experiment,
            data_loader_config=self._data_loader_config,
            data=data,
        )
        experiment_data = self._set_and_filter_training_data(
            experiment_data=experiment_data, search_space=self._model_space
        )
        # This ensures that SQ is up to date when we re-fit the existing Adapter
        # in GeneratorSpec.fit.
        self._set_status_quo(experiment=experiment)
        return self._transform_data(
            experiment_data=experiment_data,
            search_space=self._model_space,
            transforms=self._raw_transforms,
            transform_configs=self._transform_configs,
        )

    def _transform_data(
        self,
        experiment_data: ExperimentData,
        search_space: SearchSpace,
        transforms: Sequence[type[Transform]] | None,
        transform_configs: Mapping[str, TConfig],
        assign_transforms: bool = True,
    ) -> tuple[ExperimentData, SearchSpace]:
        """Initialize transforms and apply them to provided data."""
        search_space = search_space.clone()
        if transforms is not None:
            for t in transforms:
                t_instance = t(
                    search_space=search_space.clone(),
                    experiment_data=experiment_data,
                    adapter=self,
                    config=transform_configs.get(t.__name__, None),
                )
                search_space = t_instance.transform_search_space(
                    search_space=search_space
                )
                experiment_data = t_instance.transform_experiment_data(
                    experiment_data=experiment_data
                )
                if assign_transforms:
                    self.transforms[t.__name__] = t_instance
        return experiment_data, search_space

    def _set_and_filter_training_data(
        self, experiment_data: ExperimentData, search_space: SearchSpace
    ) -> ExperimentData:
        """Store non-transformed training data, and return it after filtering to
        include only in-design points if
        ``self._data_loader_config._fit_out_of_design=True``.
        """
        # NOTE: This is copied in get_training_data, so it won't be modified in-place.
        self._training_data = experiment_data
        self._metric_names: set[str] = set(experiment_data.metric_names)
        # Filter out-of-design points if `fit_out_of_design` is False.
        if self._data_loader_config.fit_out_of_design:
            self._training_in_design_idx = [True] * len(experiment_data.arm_data)
        else:
            self._training_in_design_idx = self._compute_in_design(
                search_space=search_space, experiment_data=experiment_data
            )
        return self.get_training_data(
            filter_in_design=self._data_loader_config.fit_out_of_design
        )

    def _compute_in_design(
        self,
        search_space: SearchSpace,
        experiment_data: ExperimentData,
    ) -> list[bool]:
        """Compute in-design status for each row of ``experiment_data``, after
        filling missing values if ``FillMissingParameters`` transform is used.
        """
        if "FillMissingParameters" in self._transform_configs:
            t = FillMissingParameters(
                config=self._transform_configs["FillMissingParameters"]
            )
            experiment_data = t.transform_experiment_data(
                experiment_data=experiment_data,
            )
        # TODO [T230585235]: Implement more efficient membership checks.
        return [
            search_space.check_membership(
                parameterization={k: v for k, v in params.items() if not pd.isnull(v)}
            )
            for params in experiment_data.arm_data.drop(
                # Ignoring errors which can be raised by missing metadata column
                # when the data is empty.
                columns=["metadata"],
                inplace=False,
                errors="ignore",
            ).to_dict(orient="records")
        ]

    def _set_model_space(self, arm_data: DataFrame) -> None:
        """Set model space, possibly expanding range parameters to cover data."""
        # If fill for missing values, include those in expansion.
        fill_values: TParameterization | None = None
        if "FillMissingParameters" in self._transform_configs:
            fill_values = self._transform_configs[  # pyre-ignore[9]
                "FillMissingParameters"
            ].get("fill_values", None)
        # Update model space. Expand bounds as needed to cover the values found
        # in the data. Only applies to range parameters.
        for p_name, p in self._model_space.parameters.items():
            if not isinstance(p, RangeParameter):
                continue
            if p_name in arm_data:
                param_vals = arm_data[p_name].dropna().tolist()
            else:
                param_vals = []
            if fill_values is not None and p_name in fill_values:
                param_vals.append(fill_values[p_name])
            if len(param_vals) == 0:
                continue
            p.lower = min(p.lower, min(param_vals))
            p.upper = max(p.upper, max(param_vals))
        # Remove parameter constraints from the model space.
        self._model_space.set_parameter_constraints([])

    def _set_status_quo(self, experiment: Experiment) -> None:
        """Set the status quo by extracting it from the experiment.
        The ``experiment.status_quo`` is an Arm that contains the parameterization
        and the name of the status quo arm. This method extracts the target
        trial index from the experiment, then matches the parameterization and
        trial index to the training data to make a status quo ``Observation``,
        complete with the parameterization, trial index, and data.

        NOTE: The status quo will not be set if the target trial index is None.
        If there are multiple observations for the status quo arm in the training
        data for the target trial index, we check for the map keys of the map
        metrics in the optimization config. If there is a single map key, the
        observation with the maximal map key value in the metadata will be used.
        If there are multiple map keys, the status quo will not be set.

        Args:
            experiment: The experiment to extract the status quo from.
        """
        self._status_quo: Observation | None = None  # reset the SQ.
        status_quo_arm = experiment.status_quo
        if status_quo_arm is None:
            self._status_quo_name = None
            return
        self._status_quo_name = status_quo_arm.name
        target_trial_index = get_target_trial_index(experiment=experiment)
        if target_trial_index is None:
            logger.warning(
                f"Status quo {self._status_quo_name} is not present in the "
                "training data."
            )
            return
        # Filter the training data to find the observations for status quo.
        sq_arm_observations = self._training_data.filter_by_arm_names(
            arm_names=[none_throws(self.status_quo_name)]
        ).convert_to_list_of_observations()
        status_quo_observations = [
            obs
            for obs in sq_arm_observations
            if obs.features.trial_index == target_trial_index
        ]
        if len(status_quo_observations) == 0:
            logger.warning(
                f"Status quo {self._status_quo_name} is not present in the "
                "training data."
            )
            return
        elif len(status_quo_observations) > 1:
            if self._optimization_config is None:
                logger.warning(
                    f"Status quo {self._status_quo_name} was found in the data with "
                    "multiple observations, and the Adapter does not have an "
                    "optimization config. `Adapter.status_quo` will not be set."
                )
                return

            if has_map_metrics(optimization_config=self._optimization_config):
                self._status_quo = _combine_multiple_status_quo_observations(
                    status_quo_observations=status_quo_observations,
                    metrics=set(none_throws(self._optimization_config).metrics),
                )
            else:
                logger.warning(
                    f"Status quo {self._status_quo_name} was found in the data with "
                    "multiple observations, and the optimization config does not "
                    "include any MapMetrics. `Adapter.status_quo` will not be "
                    "set."
                )
        else:
            self._status_quo = status_quo_observations[-1]

    @property
    def status_quo_data_by_trial(self) -> dict[int, ObservationData] | None:
        """A map of trial index to the status quo observation data of each trial.

        If status quo does not exist, return None.
        """
        # TODO: We could possibly extract from the experiment directly. See D72685267.
        # Status quo name will be set if status quo exists. We can just filter by name.
        if self.status_quo_name is None:
            return None
        # Identify status quo data by arm name.
        obs_data = self._training_data.observation_data
        sq_data = obs_data.loc[
            obs_data.index.get_level_values("arm_name") == self.status_quo_name
        ]
        metric_names = list(sq_data["mean"].columns)
        return {
            index[0]: ObservationData(
                metric_names=metric_names,
                means=row["mean"].to_numpy(),
                covariance=np.diag(row["sem"].to_numpy() ** 2),
            )
            for index, row in sq_data.iterrows()
        }

    @property
    def status_quo(self) -> Observation | None:
        """Observation corresponding to status quo, if any."""
        return self._status_quo

    @property
    def status_quo_name(self) -> str | None:
        """Name of status quo, if any."""
        return self._status_quo_name

    @property
    def metric_names(self) -> set[str]:
        """Metric names present in training data."""
        return self._metric_names

    @property
    def model_space(self) -> SearchSpace:
        """SearchSpace used to fit model."""
        return self._model_space

    def get_training_data(self, filter_in_design: bool = False) -> ExperimentData:
        """A copy of the (untransformed) data with which the generator was fit.

        Args:
            filter_in_design: If True, the data is filtered by
                ``self.training_in_design``. Note that this will include all
                points if ``self._data_loader_config.fit_out_of_design is True``,
                since all points will be marked as in-design.
        """
        experiment_data = deepcopy(self._training_data)
        if not filter_in_design or self._data_loader_config.fit_out_of_design:
            return experiment_data
        arm_data = experiment_data.arm_data.loc[self.training_in_design]
        obs_data = experiment_data.observation_data
        # In-design-ness is determined by the arm parameterization. So, we can
        # just filter out the observation data by the arms that are in-design.
        # We can't just use `self.training_in_design`, since there may be multiple
        # rows of observations for the same arm.
        obs_data = obs_data[
            obs_data.index.get_level_values("arm_name").isin(
                arm_data.index.get_level_values("arm_name")
            )
        ]
        return ExperimentData(arm_data=arm_data, observation_data=obs_data)

    @property
    def training_in_design(self) -> list[bool]:
        """For each observation in the training data, a bool indicating if it
        is in-design for the generator.
        """
        return self._training_in_design_idx

    def _fit(
        self,
        search_space: SearchSpace,
        experiment_data: ExperimentData,
    ) -> None:
        """Apply terminal transform and fit the generator."""
        raise AdapterMethodNotImplementedError(
            f"{self.__class__.__name__} does not implement `_fit`."
        )

    def _predict_observation_data(
        self,
        observation_features: list[ObservationFeatures],
        use_posterior_predictive: bool = False,
        untransform: bool = True,
    ) -> list[ObservationData]:
        """
        Like 'predict' method, but returns results as a list of ObservationData

        Args:
            observation_features: A list of observation features to predict.
            use_posterior_predictive: A boolean indicating if the predictions
                should be from the posterior predictive (i.e. including
                observation noise).
                This option is only supported by the ``BoTorchGenerator``.
            untransform: Whether to untransform the predictions to the original
                scale before returning.

        Returns:
            List of `ObservationData`, each representing (independent) predictions
            for the corresponding `ObservationFeatures` input.
        """
        input_len = len(observation_features)
        observation_features = deepcopy(observation_features)
        # Transform
        for t in self.transforms.values():
            observation_features = t.transform_observation_features(
                observation_features=observation_features
            )
        # Apply terminal transform and predict
        observation_data = self._predict(
            observation_features=observation_features,
            use_posterior_predictive=use_posterior_predictive,
        )
        if untransform:
            # Apply reverse transforms, in reverse order
            pred_observations = recombine_observations(
                observation_features=observation_features,
                observation_data=observation_data,
            )

            for t in reversed(list(self.transforms.values())):
                pred_observations = t.untransform_observations(pred_observations)
            observation_data = [obs.data for obs in pred_observations]
        if (output_len := len(observation_data)) != input_len:
            raise ModelError(
                f"Predictions resulted in fewer outcomes ({output_len}) than "
                f"expected ({input_len}). This can happen if a transform modifies "
                "the number of observation features, such as `Cast` dropping any "
                "observation features with `None` parameter values. "
            )
        return observation_data

    def predict(
        self,
        observation_features: list[ObservationFeatures],
        use_posterior_predictive: bool = False,
    ) -> TModelPredict:
        """Make model predictions (mean and covariance) for the given
        observation features.

        Predictions are made for all outcomes.
        If an out-of-design observation can successfully be transformed,
        the predicted value will be returned.
        Otherwise, we will attempt to find that observation in the training data
        and return the raw value.

        Args:
            observation_features: A list of observation features to predict.
            use_posterior_predictive: A boolean indicating if the predictions
                should be from the posterior predictive (i.e. including
                observation noise).
                This option is only supported by the ``BoTorchGenerator``.

        Returns:
            2-element tuple containing

            - Dictionary from metric name to list of mean estimates, in same
              order as observation_features.
            - Nested dictionary with cov['metric1']['metric2'] a list of
              cov(metric1@x, metric2@x) for x in observation_features.
        """
        # Make sure that input is a list of ObservationFeatures. If you pass in
        # arms, the code runs but it doesn't apply the transforms.
        if not all(
            isinstance(obsf, ObservationFeatures) for obsf in observation_features
        ):
            raise UserInputError(
                "Input to predict must be a list of `ObservationFeatures`."
            )
        observation_data = self._predict_observation_data(
            observation_features=observation_features,
            use_posterior_predictive=use_posterior_predictive,
        )
        f, cov = unwrap_observation_data(observation_data)
        return f, cov

    def _predict(
        self,
        observation_features: list[ObservationFeatures],
        use_posterior_predictive: bool = False,
    ) -> list[ObservationData]:
        """Apply terminal transform, predict, and reverse terminal transform on
        output.
        """
        raise AdapterMethodNotImplementedError(
            f"{self.__class__.__name__} does not implement `_predict`."
        )

    def _get_transformed_gen_args(
        self,
        search_space: SearchSpace,
        optimization_config: OptimizationConfig | None = None,
        pending_observations: dict[str, list[ObservationFeatures]] | None = None,
        fixed_features: ObservationFeatures | None = None,
    ) -> BaseGenArgs:
        if pending_observations is None:
            pending_observations = {}
        if optimization_config is None:
            optimization_config = (
                self._optimization_config.clone()
                if self._optimization_config is not None
                else None
            )
        else:
            if not self._fit_tracking_metrics:
                # Check that the optimization config has the same metrics as
                # the original one. Otherwise, we may attempt to optimize over
                # metrics that do not have a fitted model.
                outcomes = set(optimization_config.metrics.keys())
                if not outcomes.issubset(self.outcomes):
                    raise UnsupportedError(
                        "When fit_tracking_metrics is False, the optimization config "
                        "can only include metrics that were included in the "
                        "optimization config used while initializing the Adapter. "
                        f"Metrics {outcomes} is not a subset of {self.outcomes}."
                    )
            optimization_config = optimization_config.clone()

        pending_observations = deepcopy(pending_observations)
        fixed_features = (
            ObservationFeatures(parameters={})
            if fixed_features is None
            else fixed_features.clone()
        )
        search_space = search_space.clone()

        # Transform
        for t in self.transforms.values():
            search_space = t.transform_search_space(search_space)
            if optimization_config is not None:
                optimization_config = t.transform_optimization_config(
                    optimization_config=optimization_config,
                    adapter=self,
                    fixed_features=fixed_features,
                )
            for metric, po in pending_observations.items():
                pending_observations[metric] = t.transform_observation_features(po)
            if not isinstance(t, FillMissingParameters):
                (fixed_features,) = t.transform_observation_features([fixed_features])
        return BaseGenArgs(
            search_space=search_space,
            optimization_config=optimization_config,
            pending_observations=pending_observations,
            fixed_features=fixed_features,
        )

    def _validate_gen_inputs(
        self,
        n: int,
        search_space: SearchSpace | None = None,
        optimization_config: OptimizationConfig | None = None,
        pending_observations: dict[str, list[ObservationFeatures]] | None = None,
        fixed_features: ObservationFeatures | None = None,
        model_gen_options: TConfig | None = None,
    ) -> None:
        """Validate inputs to `Adapter.gen`.

        Currently, this is only used to ensure that `n` is a positive integer.
        """
        if n < 1:
            raise UserInputError(
                f"Attempted to generate n={n} points. Number of points to generate "
                "must be a positive integer."
            )

    def gen(
        self,
        n: int,
        search_space: SearchSpace | None = None,
        optimization_config: OptimizationConfig | None = None,
        pending_observations: dict[str, list[ObservationFeatures]] | None = None,
        fixed_features: ObservationFeatures | None = None,
        model_gen_options: TConfig | None = None,
    ) -> GeneratorRun:
        """
        Generate new points from the underlying generator according to
        search_space, optimization_config and other parameters.

        Args:
            n: Number of points to generate
            search_space: Search space
            optimization_config: Optimization config
            pending_observations: A map from metric name to pending
                observations for that metric.
            fixed_features: An ObservationFeatures object containing any
                features that should be fixed at specified values during
                generation.
            model_gen_options: A config dictionary that is passed along to the
                generator. See `TorchOptConfig` for details.

        Returns:
            A GeneratorRun object that contains the generated points and other metadata.
        """
        self._validate_gen_inputs(
            n=n,
            search_space=search_space,
            optimization_config=optimization_config,
            pending_observations=pending_observations,
            fixed_features=fixed_features,
            model_gen_options=model_gen_options,
        )
        t_gen_start = time.monotonic()
        # Get modifiable versions
        if search_space is None:
            search_space = self._search_space
        orig_search_space = search_space.clone()
        base_gen_args = self._get_transformed_gen_args(
            search_space=search_space,
            optimization_config=optimization_config,
            pending_observations=pending_observations,
            fixed_features=fixed_features,
        )
        # Apply terminal transform and gen
        gen_results = self._gen(
            n=n,
            search_space=base_gen_args.search_space,
            optimization_config=base_gen_args.optimization_config,
            pending_observations=base_gen_args.pending_observations,
            fixed_features=base_gen_args.fixed_features,
            model_gen_options=model_gen_options,
        )

        observation_features = gen_results.observation_features
        best_obsf = gen_results.best_observation_features

        # Apply reverse transforms
        for t in reversed(list(self.transforms.values())):
            observation_features = t.untransform_observation_features(
                observation_features
            )
            if best_obsf is not None:
                best_obsf = t.untransform_observation_features([best_obsf])[0]

        # Clamp the untransformed data to the original search space if
        # we don't fit/gen OOD points
        if not self._data_loader_config.fit_out_of_design:
            observation_features = clamp_observation_features(
                observation_features, orig_search_space
            )
            if best_obsf is not None:
                best_obsf = clamp_observation_features([best_obsf], orig_search_space)[
                    0
                ]
        best_point_predictions = None
        try:
            model_predictions = self.predict(observation_features)
            if best_obsf is not None:
                best_point_predictions = extract_arm_predictions(
                    model_predictions=self.predict([best_obsf]), arm_idx=0
                )
        except Exception as e:
            logger.debug(f"Generator predictions failed with error {e}.")
            model_predictions = None

        if best_obsf is None:
            best_arm = None
        else:
            best_arms, _ = gen_arms(
                observation_features=[best_obsf],
                arms_by_signature=self._arms_by_signature,
            )
            best_arm = best_arms[0]

        arms, candidate_metadata = gen_arms(
            observation_features=observation_features,
            arms_by_signature=self._arms_by_signature,
        )

        # If experiment has immutable search space and metrics, no need to
        # save them on generator runs.
        immutable = getattr(
            self, "_experiment_has_immutable_search_space_and_opt_config", False
        )
        optimization_config = None if immutable else base_gen_args.optimization_config
        # Remove information about the objective thresholds - we do not want to save
        # these as `ObjectiveThreshold` objects, as this causes storage headaches.
        gen_metadata = gen_results.gen_metadata
        gen_metadata.pop("objective_thresholds", None)

        generator_run = GeneratorRun(
            arms=arms,
            weights=gen_results.weights,
            optimization_config=optimization_config,
            search_space=None if immutable else orig_search_space,
            model_predictions=model_predictions,
            best_arm_predictions=(
                None if best_arm is None else (best_arm, best_point_predictions)
            ),
            fit_time=self.fit_time_since_gen,
            gen_time=time.monotonic() - t_gen_start,
            model_key=self._model_key,
            model_kwargs=self._model_kwargs,
            bridge_kwargs=self._bridge_kwargs,
            gen_metadata=gen_metadata,
            model_state_after_gen=self._get_serialized_model_state(),
            candidate_metadata_by_arm_signature=candidate_metadata,
        )
        if len(generator_run.arms) < n:
            logger.warning(
                f"{self} was not able to generate {n} unique candidates. "
                "Generated arms have the following weights, as there are repeats:\n"
                f"{generator_run.weights}"
            )
        self.fit_time_since_gen = 0.0
        return generator_run

    def _gen(
        self,
        n: int,
        search_space: SearchSpace,
        optimization_config: OptimizationConfig | None,
        pending_observations: dict[str, list[ObservationFeatures]],
        fixed_features: ObservationFeatures | None,
        model_gen_options: TConfig | None,
    ) -> GenResults:
        """Apply terminal transform, gen, and reverse terminal transform on
        output.
        """
        raise AdapterMethodNotImplementedError(
            f"{self.__class__.__name__} does not implement `_gen`."
        )

    def cross_validate(
        self,
        cv_training_data: ExperimentData,
        cv_test_points: list[ObservationFeatures],
        use_posterior_predictive: bool = False,
    ) -> list[ObservationData]:
        """Make a set of cross-validation predictions.

        Args:
            cv_training_data: The training data to use for cross validation.
            cv_test_points: The test points at which predictions will be made.
            use_posterior_predictive: A boolean indicating if the predictions
                should be from the posterior predictive (i.e. including
                observation noise).

        Returns:
            A list of predictions at the test points.
        """
        # Apply transforms to cv_training_data and cv_test_points
        cv_training_data, cv_test_points, search_space = self._transform_inputs_for_cv(
            cv_training_data=cv_training_data, cv_test_points=cv_test_points
        )

        # Apply terminal transform, and get predictions.
        # Since each CV fold removes points from the training data, the
        # remaining observations will not pass the input scaling checks.
        # To avoid confusing users with warnings, we disable these checks.
        with validate_input_scaling(False):
            cv_predictions = self._cross_validate(
                search_space=search_space,
                cv_training_data=cv_training_data,
                cv_test_points=cv_test_points,
                use_posterior_predictive=use_posterior_predictive,
            )
        # Apply reverse transforms, in reverse order
        cv_test_observations = [
            Observation(features=obsf, data=cv_predictions[i])
            for i, obsf in enumerate(cv_test_points)
        ]

        for t in reversed(list(self.transforms.values())):
            cv_test_observations = t.untransform_observations(cv_test_observations)
        return [obs.data for obs in cv_test_observations]

    def _cross_validate(
        self,
        search_space: SearchSpace,
        cv_training_data: ExperimentData,
        cv_test_points: list[ObservationFeatures],
        use_posterior_predictive: bool = False,
    ) -> list[ObservationData]:
        """Apply the terminal transform, make predictions on the test points,
        and reverse terminal transform on the results.
        """
        raise AdapterMethodNotImplementedError(
            f"{self.__class__.__name__} does not implement `_cross_validate`."
        )

    def _transform_inputs_for_cv(
        self,
        cv_training_data: ExperimentData,
        cv_test_points: list[ObservationFeatures],
    ) -> tuple[ExperimentData, list[ObservationFeatures], SearchSpace]:
        """Apply transforms to cv_training_data and cv_test_points,
        and return cv_training_data, cv_test_points, and search space in
        transformed space. This is to prepare data to be used in _cross_validate.

        Args:
            cv_training_data: The training data to use for cross validation.
            cv_test_points: The test points at which predictions will be made.

        Returns:
            cv_training_data, cv_test_points, and search_space in transformed space.
        """
        cv_test_points = deepcopy(cv_test_points)
        cv_training_data = deepcopy(cv_training_data)
        search_space = self._model_space.clone()
        for t in self.transforms.values():
            cv_training_data = t.transform_experiment_data(
                experiment_data=cv_training_data
            )
            cv_test_points = t.transform_observation_features(
                observation_features=cv_test_points
            )
            search_space = t.transform_search_space(search_space=search_space)
        return cv_training_data, cv_test_points, search_space

    def _set_kwargs_to_save(
        self,
        model_key: str,
        model_kwargs: dict[str, Any],
        bridge_kwargs: dict[str, Any],
    ) -> None:
        """Set properties used to save the model that created a given generator
        run, on the `GeneratorRun` object. Each generator run produced by the
        `gen` method of this adapter will have the model key and kwargs
        fields set as provided in arguments to this function.
        """
        self._model_key = model_key
        self._model_kwargs = model_kwargs
        self._bridge_kwargs = bridge_kwargs

    def _get_serialized_model_state(self) -> dict[str, Any]:
        """Obtains the state of the underlying generator (if using a stateful one)
        in a readily JSON-serializable form.
        """
        return self.generator.serialize_state(raw_state=self.generator._get_state())

    def _deserialize_model_state(
        self, serialized_state: dict[str, Any]
    ) -> dict[str, Any]:
        return self.generator.deserialize_state(serialized_state=serialized_state)

    def feature_importances(self, metric_name: str) -> dict[str, float]:
        """Computes feature importances for a single metric.

        Depending on the type of the generator, this method will approach sensitivity
        analysis (calculating the sensitivity of the metric to changes in the search
        space's parameters, a.k.a. features) differently.

        For Bayesian optimization generators (BoTorch generators), this method uses
        parameter inverse lengthscales to compute normalized feature importances.

        NOTE: Currently, this is only implemented for GP-based generators.

        Args:
            metric_name: Name of metric to compute feature importances for.

        Returns:
            A dictionary mapping parameter names to their corresponding feature
            importances.

        """
        raise AdapterMethodNotImplementedError(
            f"{self.__class__.__name__} does not implement `feature_importances`."
        )

    def transform_observations(self, observations: list[Observation]) -> Any:
        """Applies transforms to given observation features and returns them in the
        model space.

        Args:
            observation_features: ObservationFeatures to be transformed.

        Returns:
            Transformed values. This could be e.g. a torch Tensor, depending
            on the Adapter subclass.
        """
        observations = deepcopy(observations)
        for t in self.transforms.values():
            observations = t.transform_observations(observations)
        # Apply terminal transform and return
        return self._transform_observations(observations)

    def _transform_observations(self, observations: list[Observation]) -> Any:
        """Apply terminal transform to given observations and return result."""
        raise AdapterMethodNotImplementedError(
            f"{self.__class__.__name__} does not implement `_transform_observations`."
        )

    def transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> Any:
        """Applies transforms to given observation features and returns them in the
        model space.

        Args:
            observation_features: ObservationFeatures to be transformed.

        Returns:
            Transformed values. This could be e.g. a torch Tensor, depending
            on the Adapter subclass.
        """
        obsf = deepcopy(observation_features)
        for t in self.transforms.values():
            obsf = t.transform_observation_features(obsf)
        # Apply terminal transform and return
        return self._transform_observation_features(obsf)

    def _transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> Any:
        """Apply terminal transform to given observation features and return result."""
        raise AdapterMethodNotImplementedError(
            f"{self.__class__.__name__} does not implement "
            "`_transform_observation_features`."
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(generator={self.generator})"


def unwrap_observation_data(observation_data: list[ObservationData]) -> TModelPredict:
    """Converts observation data to the format for model prediction outputs.
    That format assumes each observation data has the same set of metrics.
    """
    metrics = set(observation_data[0].metric_names)
    f: TModelMean = {metric: [] for metric in metrics}
    cov: TModelCov = {m1: {m2: [] for m2 in metrics} for m1 in metrics}
    for od in observation_data:
        if set(od.metric_names) != metrics:
            raise ValueError(
                "Each ObservationData should use same set of metrics. "
                "Expected {exp}, got {got}.".format(
                    exp=metrics, got=set(od.metric_names)
                )
            )
        for i, m1 in enumerate(od.metric_names):
            f[m1].append(od.means[i])
            for j, m2 in enumerate(od.metric_names):
                cov[m1][m2].append(od.covariance[i, j])
    return f, cov


def gen_arms(
    observation_features: list[ObservationFeatures],
    arms_by_signature: dict[str, Arm] | None = None,
) -> tuple[list[Arm], dict[str, TCandidateMetadata] | None]:
    """Converts observation features to a tuple of arms list and candidate metadata
    dict, where arm signatures are mapped to their respective candidate metadata.
    """
    # TODO(T34225939): handle static context (which is stored on observation_features)
    arms = []
    candidate_metadata = {}
    for of in observation_features:
        arm = Arm(parameters=of.parameters)
        if arms_by_signature is not None and arm.signature in arms_by_signature:
            existing_arm = arms_by_signature[arm.signature]
            arm = Arm(name=existing_arm.name, parameters=existing_arm.parameters)
        arms.append(arm)
        if of.metadata:
            candidate_metadata[arm.signature] = of.metadata
    return arms, candidate_metadata or None  # None if empty cand. metadata.


def clamp_observation_features(
    observation_features: list[ObservationFeatures], search_space: SearchSpace
) -> list[ObservationFeatures]:
    range_parameters = [
        p for p in search_space.parameters.values() if isinstance(p, RangeParameter)
    ]
    for obsf in observation_features:
        for p in range_parameters:
            if p.name not in obsf.parameters:
                continue
            if p.parameter_type == ParameterType.FLOAT:
                val = assert_is_instance(obsf.parameters[p.name], float)
            else:
                val = assert_is_instance(obsf.parameters[p.name], int)
            if val < p.lower:
                logger.debug(
                    f"Untransformed parameter {val} "
                    f"less than lower bound {p.lower}, clamping"
                )
                obsf.parameters[p.name] = p.lower
            elif val > p.upper:
                logger.debug(
                    f"Untransformed parameter {val} "
                    f"greater than upper bound {p.upper}, clamping"
                )
                obsf.parameters[p.name] = p.upper
    return observation_features


def _legacy_overwrite_data_loader_config(
    data_loader_config: DataLoaderConfig,
    fit_out_of_design: bool | None = None,
    fit_abandoned: bool | None = None,
    fit_only_completed_map_metrics: bool | None = None,
    warn_if_legacy: bool = True,
) -> DataLoaderConfig:
    """Overwrites data loader config with legacy keyword arguments.

    Args:
        data_loader_config: Data loader config.
        fit_out_of_design: Whether to fit out-of-design points.
        fit_abandoned: Whether to fit abandoned arms.
        fit_only_completed_map_metrics: Whether to fit only completed map metrics.
        warn_if_legacy: Whether to warn if legacy keyword arguments are used.

    Returns:
        Updated data loader config.
    """
    data_loader_config_dict = {}
    for var_name, deprecated_var in (
        ("fit_out_of_design", fit_out_of_design),
        ("fit_abandoned", fit_abandoned),
        ("fit_only_completed_map_metrics", fit_only_completed_map_metrics),
    ):
        if deprecated_var is not None:
            if warn_if_legacy:
                logger.warning(
                    f"`{var_name}` is deprecated. Please pass as "
                    f"`data_loader_config.{var_name}` instead."
                )
            data_loader_config_dict[var_name] = deprecated_var
        else:
            data_loader_config_dict[var_name] = getattr(data_loader_config, var_name)

    data_loader_config = DataLoaderConfig(
        latest_rows_per_group=data_loader_config.latest_rows_per_group,
        limit_rows_per_metric=data_loader_config.limit_rows_per_metric,
        limit_rows_per_group=data_loader_config.limit_rows_per_group,
        **data_loader_config_dict,
    )
    return data_loader_config


def _combine_multiple_status_quo_observations(
    status_quo_observations: list[Observation],
    metrics: set[str],
) -> Observation | None:
    """Finds the maximal (in terms of map key value) observation for each metric
    in `status_quo_observations`, and combines them into a single ``Observation``
    object, representing the status quo observations for all metrics.

    NOTE: The resulting ``ObservationFeatures`` will not have any ``metadata``.
    If there are multiple ``Observation``s for the status quo, this is due to
    them having different ``metadata``, so we discard it here to avoid having
    misleading or incomplete information for some metrics.

    Args:
        status_quo_observations: List of observations for the status quo
            arm at target trial index. Extracted in ``Adapter._set_status_quo``.
        metrics: The metrics to include in the combined observation.
            This should include all metrics in the optimization config.

    Returns:
        A single ``Observation`` object that includes the maximal observation
        for each metric.
    """
    # Pick the observation with maximal map key value.
    partial_obs = [
        max(
            status_quo_observations,
            key=lambda obs: obs.features.metadata[MAP_KEY],
        )
    ]
    # Check if the it includes all metrics in the opt config.
    # If not, search for observations of the remaining metrics as well.
    while remaining_metrics := metrics.difference(
        sum((obs.data.metric_names for obs in partial_obs), [])
    ):
        # Find observations of the remaining metrics.
        # Search using one metric at a time.
        lookup_metric = remaining_metrics.pop()
        obs_w_lookup_metric = [
            obs
            for obs in status_quo_observations
            if lookup_metric in obs.data.metric_names
        ]
        if len(obs_w_lookup_metric) == 0:
            logger.warning(
                f"Could not find observations of metric {lookup_metric} for the "
                f"status quo {status_quo_observations[0].arm_name} in the training "
                "data. `Adapter.status_quo` will not be set."
            )
            return
        partial_obs.append(
            max(
                obs_w_lookup_metric,
                key=lambda obs: obs.features.metadata[MAP_KEY],
            )
        )
    # Combine into a single Observation object.
    return Observation(
        features=ObservationFeatures(
            parameters=partial_obs[0].features.parameters,
            trial_index=partial_obs[0].features.trial_index,
            # NOTE: omitting the metadata since it can be different in each obs.
        ),
        data=ObservationData(
            metric_names=sum((obs.data.metric_names for obs in partial_obs), []),
            means=np.concatenate([obs.data.means for obs in partial_obs], axis=0),
            covariance=np.diag(
                np.concatenate(
                    [np.diag(obs.data.covariance) for obs in partial_obs], axis=0
                )
            ),
        ),
        arm_name=partial_obs[0].arm_name,
    )
